#  Copyright (c) 2021. Robin Thibaut, Ghent University
import os
from datetime import date
from os.path import join as jp

from loguru import logger

__version__ = "1.0.dev0"

source = __name__.split(".")[-1]
# Set up logger
logger.add(
    jp(os.getcwd(), "logs", f"{source}_{date.today()}.log"),
    backtrace=True,
    diagnose=True,
    enqueue=True,
)

__all__ = [
    "utils",
    "goggles",
    "algorithms",
    "learning",
    "preprocessing",
    "spatial",
]



#  Copyright (c) 2021. Robin Thibaut, Ghent University
from typing import List

import numpy as np
from matplotlib import pyplot as plt

plt.rcParams.update({"figure.max_open_warning": 0})

__all__ = [
    "grid_parameters",
    "block_shaped",
    "refine_axis",
    "rc_from_blocks",
    "blocks_from_rc",
    "blocks_from_rc_3d",
    "get_centroids",
    "contour_extract",
    "contours_vertices",
    "refine_machine",
]


def grid_parameters(
    x_lim: list = None, y_lim: list = None, grf: float = 1
) -> (np.array, int, int):
    """
    Generates grid parameters given dimensions.
    :param x_lim:
    :param y_lim:
    :param grf:
    :return:
    """
    if y_lim is None:
        y_lim = [0, 1000]
    else:
        y_lim = y_lim
    if x_lim is None:
        x_lim = [0, 1500]
    else:
        x_lim = x_lim

    grf = grf  # Cell dimension
    nrow = int(np.diff(y_lim) / grf)  # Number of rows
    ncol = int(np.diff(x_lim) / grf)  # Number of columns
    array = np.ones((nrow, ncol))  # Dummy array
    # Centroids of dummy array
    xys = get_centroids(array, grf) + np.min([x_lim, y_lim], axis=1)

    return xys, nrow, ncol


def block_shaped(arr: np.array, nrows: int, ncols: int) -> np.array:
    """
    Return an array of shape (n, nrows, ncols) where
    n * nrows * ncols = arr.size

    If arr is a 2D array, the returned array should look like n sub-blocks with
    each sub-block preserving the "physical" layout of arr.
    """
    h, w = arr.shape
    assert h % nrows == 0, "{} rows is not evenly divisible by {}".format(h, nrows)
    assert w % ncols == 0, "{} cols is not evenly divisible by {}".format(w, ncols)

    return (
        arr.reshape(h // nrows, nrows, -1, ncols)
        .swapaxes(1, 2)
        .reshape(-1, nrows, ncols)
    )


def refine_axis(
    widths: List[float], r_pt: float, ext: float, cnd: float, d_dim: float, a_lim: float
) -> np.array:
    """
    Refines one 1D axis around a point belonging to it.

    Example:
    along_c = refine_axis([10m, 10m... 10m], 500m, 70m, 2m, 10m, 1500m)

    :param widths: Array of cell widths along axis.
    :param r_pt: 1D point on the axis around which refining will occur.
    :param ext: Extent (distance) of the refinement around the point.
    :param cnd: New cell size after refinement.
    :param d_dim: Base cell dimension.
    :param a_lim: Limit of the axis.
    :return: Refined axis (widths)
    """

    x0 = widths
    x0s = np.cumsum(x0)  # Cumulative sum of the width of the cells
    pt = r_pt  # Point around which refining
    extx = ext  # Extent around the point
    cdrx = cnd
    dx = d_dim
    xlim = a_lim

    # X range of the polygon
    xrp = [pt - extx, pt + extx]

    # Where to refine
    wherex = np.where((xrp[0] < x0s) & (x0s <= xrp[1]))[0]

    # The algorithm must choose a 'flexible parameter', either the cell grid size, the dimensions of the grid or the
    # refined cells themselves... We choose to adapt the dimensions of the grid.
    exn = np.sum(x0[wherex])  # x-extent of the refinement zone
    fx = exn / cdrx  # divides the extent by the new cell spacing
    rx = exn % cdrx  # remainder
    if rx == 0:
        nwxs = np.ones(int(fx)) * cdrx
        x0 = np.delete(x0, wherex)
        x0 = np.insert(x0, wherex[0], nwxs)
    else:  # If the cells can not be exactly subdivided into the new cell dimension
        nwxs = np.ones(int(round(fx))) * cdrx  # Produce a new width vector
        x0 = np.delete(x0, wherex)  # Delete old cells
        x0 = np.insert(x0, wherex[0], nwxs)  # insert new

        # Cumulative width should equal x_lim, but it will not be the case, we have to adapt widths.
        cs = np.cumsum(x0)
        difx = xlim - cs[-1]
        # Location of cells whose widths will be adapted
        where_default = np.where(abs(x0 - dx) <= 5)[0]
        # Where do we have the default cell size on
        where_left = where_default[np.where(where_default < wherex[0])]
        # the left
        where_right = where_default[
            np.where((where_default >= wherex[0] + len(nwxs)))
        ]  # And on the right
        lwl = len(where_left)
        lwr = len(where_right)

        if lwl > lwr:
            rl = (
                lwl / lwr
            )  # Weights how many cells are on either sides of the refinement zone
            # Splitting the extra widths on the left and right of the cells
            dal = difx / ((lwl + lwr) / lwl)
            dal = dal + (difx - dal) / rl
            dar = difx - dal
        elif lwr > lwl:
            rl = (
                lwr / lwl
            )  # Weights how many cells are on either sides of the refinement zone
            # Splitting the extra widths on the left and right of the cells
            dar = difx / ((lwl + lwr) / lwr)
            dar = dar + (difx - dar) / rl
            dal = difx - dar
        else:
            # Splitting the extra widths on the left and right of the cells
            dal = difx / ((lwl + lwr) / lwl)
            dar = difx - dal

        x0[where_left] = x0[where_left] + dal / lwl
        x0[where_right] = x0[where_right] + dar / lwr

    return x0


def rc_from_blocks(blocks: np.array) -> (np.array, np.array):
    """
    Computes the x and y dimensions of each block
    :param blocks:
    :return:
    """
    dc = np.array([np.diff(b[:, 0]).max for b in blocks])
    dr = np.array([np.diff(b[:, 1]).max for b in blocks])

    return dc, dr


def blocks_from_rc_3d(rows: np.array, columns: np.array) -> np.array:
    """
    Returns the blocks forming a 2D grid whose rows and columns widths are defined by the two arrays rows, columns
    """

    nrow = len(rows)
    ncol = len(columns)
    delr = rows
    delc = columns
    r_sum = np.cumsum(delr)
    c_sum = np.cumsum(delc)

    blocks = []
    for c in range(nrow):
        for n in range(ncol):
            b = [
                [c_sum[n] - delc[n], r_sum[c] - delr[c], 0.0],
                [c_sum[n] - delc[n], r_sum[c], 0.0],
                [c_sum[n], r_sum[c], 0.0],
                [c_sum[n], r_sum[c] - delr[c], 0.0],
            ]
            blocks.append(b)
    blocks = np.array(blocks)

    return blocks


def blocks_from_rc(rows: np.array, columns: np.array) -> np.array:
    """
    Returns the blocks forming a 2D grid whose rows and columns widths are defined by the two arrays rows, columns
    """

    nrow = len(rows)
    ncol = len(columns)
    delr = rows
    delc = columns
    r_sum = np.cumsum(delr)
    c_sum = np.cumsum(delc)

    blocks = []
    for c in range(nrow):
        for n in range(ncol):
            b = [
                [c_sum[n] - delc[n], r_sum[c] - delr[c]],
                [c_sum[n] - delc[n], r_sum[c]],
                [c_sum[n], r_sum[c]],
                [c_sum[n], r_sum[c] - delr[c]],
            ]
            blocks.append(b)
    blocks = np.array(blocks)

    return blocks


def get_centroids(array: np.array, grf: float) -> np.array:
    """
    Given a (m, n) matrix of cells dimensions in the x-y axes, returns the (m, n, 2) matrix of the coordinates of
    centroids.
    :param array: (m, n) array
    :param grf: float: Cell dimension
    """
    xys = np.dstack(
        (np.flip((np.indices(array.shape) + 1), 0) * grf - grf / 2)
    )  # Getting centroids
    return xys.reshape((array.shape[0] * array.shape[1], 2))


# extract 0 contours
def contour_extract(x_lim, y_lim, grf, Z):
    """
    Extract the 0 contour from the sampled posterior, corresponding to the WHPA delineation
    """
    *_, x, y = refine_machine(x_lim, y_lim, grf)
    vertices = contours_vertices(x, y, Z)

    return x, y, vertices


def contours_vertices(
    x: list, y: list, arrays: np.array, c: float = 0, ignore_: bool = True
) -> np.array:
    """
    Extracts contour vertices from a list of matrices.
    :param x:
    :param y:
    :param arrays: list of matrices
    :param c: Contour value
    :param ignore_: Bool value to consider multiple contours or not (see comments)
    :return: vertices array
    """
    if len(arrays.shape) < 3:
        arrays = [arrays]
    # First create figures for each forecast.
    figs = [plt.figure() for _ in range(len(arrays))]
    c0s = [plt.contour(x, y, f, [c]) for f in arrays]
    [plt.close(f) for f in figs]  # Close plots
    # .allseg[0][0] extracts the vertices of each O contour = WHPA's vertices

    # /!\ If more than one contours are present for the same values, possibility to include them or not.
    # How are the contours sorted in contour_extract.allsegs[0][i] ?
    # It looks like they are sorted by size.
    if ignore_:
        v = np.array([c0.allsegs[0][0] for c0 in c0s], dtype=object)
    else:
        v = np.array(
            [c0.allsegs[0][i] for c0 in c0s for i in range(len(c0.allsegs[0]))],
            dtype=object,
        )
    return v


def refine_machine(
    xlim: list, ylim: list, new_grf: int or float
) -> (int, int, np.array, np.array):
    nrow = int(np.diff(ylim) / new_grf)  # Number of rows
    ncol = int(np.diff(xlim) / new_grf)  # Number of columns
    new_x, new_y = np.meshgrid(
        np.linspace(xlim[0], xlim[1], ncol), np.linspace(ylim[0], ylim[1], nrow)
    )
    return nrow, ncol, new_x, new_y



#  Copyright (c) 2021. Robin Thibaut, Ghent University

import os
import string
from os.path import join as jp

import joblib
import numpy as np
import pandas as pd
import seaborn as sns
from loguru import logger
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon
from scipy.interpolate import BSpline, make_interp_spline
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array, deprecated

from demo_config import Setup
from skbel import utils
from skbel.goggles import explained_variance, _proxy_annotate, _proxy_legend, _kde_cca
from skbel.spatial import (
    contours_vertices,
    grid_parameters,
    refine_machine,
)


def binary_polygon(
    xys: np.array,
    nrow: int,
    ncol: int,
    pzs: np.array,
    outside: float = -1,
    inside: float = 1,
) -> np.array:
    """
    Given a polygon whose vertices are given by the array pzs, and a matrix of
    centroids coordinates of the surface discretization, assigns to the matrix a certain value
    whether the cell is inside or outside said polygon.

    To compute the signed distance function, we need a negative/positive value.

    :param xys: Centroids of a grid' cells
    :param nrow: Number of rows
    :param ncol: Number of columns
    :param pzs: Array of ordered vertices coordinates of a polygon.
    :param pzs: Polygon vertices (v, 2)
    :param outside: Value to assign to the matrix outside of the polygon
    :param inside: Value to assign to the matrix inside of the polygon
    :return: phi = the binary matrix
    """

    # Creates a Polygon abject out of the polygon vertices in pzs
    poly = Polygon(pzs, True)
    # Checks which points are enclosed by polygon.
    ind = np.nonzero(poly.contains_points(xys))[0]
    phi = np.ones((nrow, ncol)) * outside  # SD - create matrix of 'outside'
    phi = phi.reshape((nrow * ncol))  # Flatten to have same dimension as 'ind'
    phi[ind] = inside  # Points inside the WHPA are assigned a value of 'inside'
    phi = phi.reshape((nrow, ncol))  # Reshape

    return phi


def binary_stack(xys: np.array, nrow: int, ncol: int, vertices: np.array) -> np.array:
    """
    Takes WHPA vertices and 'binarizes' the image (e.g. 1 inside, 0 outside WHPA).
    """
    # Create binary images of WHPA stored in bin_whpa
    bin_whpa = [
        binary_polygon(xys, nrow, ncol, pzs=p, inside=1, outside=-1) for p in vertices
    ]
    big_sum = np.sum(bin_whpa, axis=0)  # Stack them
    # Scale from 0 to 1
    big_sum -= np.min(big_sum)
    big_sum /= np.max(big_sum)
    return big_sum


def reload_trained_model(base_dir: str, root: str = None, well: str = None):
    if root is None:
        root = ""
    if well is None:
        well = ""

    res_dir = jp(base_dir, root, well, "obj")

    bel = joblib.load(jp(res_dir, "bel.pkl"))

    return bel


def pca_scores(
    training: np.array,
    prediction: np.array,
    n_comp: int,
    annotation: list = None,
    fig_file: str = None,
    labels: bool = True,
    show: bool = False,
):
    """
    PCA scores plot, displays scores of observations above those of training.
    :param labels:
    :param training: Training scores
    :param prediction: Test scores
    :param n_comp: How many components to show
    :param annotation: List of annotation(s)
    :param fig_file:
    :param show:
    :return:
    """
    if annotation is None:
        annotation = []
    # Scores plot
    # Grid
    plt.grid(alpha=0.2)
    # Ticks
    plt.xticks(
        np.concatenate([np.array([0]), np.arange(4, n_comp, 5)]),
        np.concatenate([np.array([1]), np.arange(5, n_comp + 5, 5)]),
    )  # Plot all training scores
    plt.plot(training.T[:n_comp], "ob", markersize=3, alpha=0.1)
    # plt.plot(training.T[:ut], '+w', markersize=.5, alpha=0.2)

    # For each sample used for prediction:
    for sample_n in range(len(prediction)):
        # Select observation
        pc_obs = prediction[sample_n]
        # Create beautiful spline to follow prediction scores
        xnew = np.linspace(1, n_comp, 200)  # New points for plotting curve
        spl = make_interp_spline(
            np.arange(1, n_comp + 1), pc_obs.T[:n_comp], k=3
        )  # type: BSpline
        power_smooth = spl(xnew)
        # I forgot why I had to put '-1'
        plt.plot(xnew - 1, power_smooth, "red", linewidth=1.2, alpha=0.9)

        plt.plot(
            pc_obs.T[:n_comp],  # Plot observations scores
            "ro",
            markersize=3,
            markeredgecolor="k",
            markeredgewidth=0.4,
            alpha=0.8,
            label=str(sample_n),
        )

    if labels:
        plt.title("Principal Components of training and test dataset")
        plt.xlabel("PC number")
        plt.ylabel("PC")
    plt.tick_params(labelsize=11)
    # Add legend
    # Add title inside the box
    legend_a = _proxy_annotate(annotation=annotation, loc=2, fz=14)
    _proxy_legend(
        legend1=legend_a,
        colors=["blue", "red"],
        labels=["Training", "Test"],
        marker=["o", "o"],
    )

    if fig_file:
        utils.dirmaker(os.path.dirname(fig_file))
        plt.savefig(fig_file, dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()


def whpa_plot(
    grf: float = None,
    well_comb: list = None,
    whpa: np.array = None,
    alpha: float = 0.4,
    halpha: float = None,
    lw: float = 0.5,
    bkg_field_array: np.array = None,
    vmin: float = None,
    vmax: float = None,
    x_lim: list = None,
    y_lim: list = None,
    xlabel: str = None,
    ylabel: str = None,
    cb_title: str = None,
    labelsize: float = 5,
    cmap: str = "coolwarm",
    color: str = "white",
    grid: bool = True,
    show_wells: bool = False,
    well_ids: list = None,
    title: str = None,
    annotation: list = None,
    fig_file: str = None,
    highlight: bool = False,
    show: bool = False,
):
    """
    Produces the WHPA plot, i.e. the zero-contour of the signed distance array.

    :param grid:
    :param grf: Grid cell size
    :param well_comb: List of well combination
    :param highlight: Boolean to display lines on top of filling between contours or not.
    :param annotation: List of annotations (str)
    :param xlabel:
    :param ylabel:
    :param cb_title:
    :param well_ids:
    :param labelsize: Label size
    :param title: str: plot title
    :param show_wells: bool: whether to plot well coordinates or not
    :param cmap: str: colormap name for the background array
    :param vmax: float: max value to plot for the background array
    :param vmin: float: max value to plot for the background array
    :param bkg_field_array: np.array: 2D array whose values will be plotted on the grid
    :param whpa: np.array: Array containing grids of values whose 0 contour will be computed and plotted
    :param alpha: float: opacity of the 0 contour lines
    :param halpha: Alpha value for line plots if highlight is True
    :param lw: float: Line width
    :param color: str: Line color
    :param fig_file: str:
    :param show: bool:
    :param x_lim: [x_min, x_max] For the figure
    :param y_lim: [y_min, y_max] For the figure
    """

    # Get basic settings
    focus = Setup.Focus
    wells = Setup.Wells

    if well_comb is not None:
        wells.combination = well_comb

    if y_lim is None:
        ylim = focus.y_range
    else:
        ylim = y_lim
    if x_lim is None:
        xlim = focus.x_range
    else:
        xlim = x_lim
    if grf is None:
        grf = focus.cell_dim
    else:
        grf = grf

    nrow, ncol, x, y = refine_machine(focus.x_range, focus.y_range, grf)

    # Plot background
    if bkg_field_array is not None:
        plt.imshow(
            bkg_field_array,
            extent=np.concatenate([focus.x_range, focus.y_range]),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
        )
        cb = plt.colorbar(fraction=0.046, pad=0.04)
        cb.ax.set_title(cb_title)

    if halpha is None:
        halpha = alpha

    # Plot results
    if whpa is None:
        whpa = []

    if whpa.ndim > 2:  # New approach is to plot filled contours
        new_grf = 1  # Refine grid
        _, _, new_x, new_y = refine_machine(xlim, ylim, new_grf=new_grf)
        xys, nrow, ncol = grid_parameters(x_lim=xlim, y_lim=ylim, grf=new_grf)
        vertices = contours_vertices(x=x, y=y, arrays=whpa)
        b_low = binary_stack(xys=xys, nrow=nrow, ncol=ncol, vertices=vertices)
        contour = plt.contourf(
            new_x,
            new_y,
            1 - b_low,  # Trick to be able to fill contours
            # Use machine epsilon
            [np.finfo(float).eps, 1 - np.finfo(float).eps],
            colors=color,
            alpha=alpha,
        )
        if highlight:  # Also display curves
            for z in whpa:
                contour = plt.contour(
                    z,
                    [0],
                    extent=(xlim[0], xlim[1], ylim[0], ylim[1]),
                    colors=color,
                    linewidths=lw,
                    alpha=halpha,
                )

    else:  # If only one WHPA to display
        contour = plt.contour(
            whpa,
            [0],
            extent=np.concatenate([focus.x_range, focus.y_range]),
            colors=color,
            linewidths=lw,
            alpha=halpha,
        )

    # Grid
    if grid:
        plt.grid(color="c", linestyle="-", linewidth=0.5, alpha=0.2)

    # Plot wells
    well_legend = None
    if show_wells:
        plot_wells(wells, well_ids=well_ids, markersize=7)
        well_legend = plt.legend(fontsize=11)

    # Plot limits
    if x_lim is None:
        plt.xlim(xlim[0], xlim[1])
    else:
        plt.xlim(x_lim[0], x_lim[1])
    if y_lim is None:
        plt.ylim(ylim[0], ylim[1])
    else:
        plt.ylim(y_lim[0], y_lim[1])

    if title:
        plt.title(title)

    plt.xlabel(xlabel, fontsize=labelsize)
    plt.ylabel(ylabel, fontsize=labelsize)

    # Tick size
    plt.tick_params(labelsize=labelsize)

    if annotation:
        legend = _proxy_annotate(annotation=annotation, fz=14, loc=2)
        plt.gca().add_artist(legend)

    if fig_file:
        utils.dirmaker(os.path.dirname(fig_file))
        plt.savefig(fig_file, bbox_inches="tight", dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()
        plt.close()

    return contour, well_legend


@deprecated()
def post_examination(
    root: str, xlim: list = None, ylim: list = None, show: bool = False
):
    focus = Setup.Focus()
    if xlim is None:
        xlim = focus.x_range
    if ylim is None:
        ylim = focus.y_range  # [335, 700]
    md = Setup.Directories()
    ndir = jp(md.forecasts_dir, "base", "roots_whpa", f"{root}.npy")
    sdir = os.path.dirname(ndir)
    nn = np.load(ndir)
    whpa_plot(
        whpa=nn,
        x_lim=xlim,
        y_lim=ylim,
        labelsize=11,
        alpha=1,
        xlabel="X(m)",
        ylabel="Y(m)",
        cb_title="SD(m)",
        annotation=["B"],
        bkg_field_array=np.flipud(nn[0]),
        color="black",
        cmap="coolwarm",
    )

    # legend = proxy_annotate(annotation=['B'], loc=2, fz=14)
    # plt.gca().add_artist(legend)

    plt.savefig(
        jp(sdir, f"{root}_SD.png"), dpi=300, bbox_inches="tight", transparent=False
    )
    if show:
        plt.show()
    plt.close()


def h_pca_inverse_plot(bel, fig_dir: str = None, show: bool = False):
    """
    Plot used to compare the reproduction of the original physical space after PCA transformation
    :param bel: BEL model
    :param fig_dir: str:
    :param show: bool:
    :return:
    """

    shape = bel.Y_shape

    if bel.Y_obs_pc is not None:
        v_pc = check_array(bel.Y_obs_pc.reshape(1, -1))
    else:
        try:
            Y_obs = check_array(bel.Y_obs)
        except ValueError:
            Y_obs = check_array(bel.Y_obs.to_numpy().reshape(1, -1))

        v_pc = bel.Y_pre_processing.transform(Y_obs)[
            :, : Setup.HyperParameters.n_pc_target
        ]

    nc = bel.Y_pre_processing["pca"].n_components_
    dummy = np.zeros((1, nc))
    dummy[:, : v_pc.shape[1]] = v_pc
    v_pred = bel.Y_pre_processing.inverse_transform(dummy)

    h_to_plot = np.copy(Y_obs.reshape(1, shape[1], shape[2]))

    whpa_plot(whpa=h_to_plot, color="red", alpha=1, lw=2)

    whpa_plot(
        whpa=v_pred.reshape(1, shape[1], shape[2]),
        color="blue",
        alpha=1,
        lw=2,
        labelsize=11,
        xlabel="X(m)",
        ylabel="Y(m)",
        x_lim=[850, 1100],
        y_lim=[350, 650],
    )

    # Add title inside the box
    an = ["B"]

    legend_a = _proxy_annotate(annotation=an, loc=2, fz=14)

    _proxy_legend(
        legend1=legend_a,
        colors=["red", "blue"],
        labels=["Physical", "Back transformed"],
        marker=["-", "-"],
    )

    if fig_dir is not None:
        utils.dirmaker(fig_dir)
        plt.savefig(
            jp(fig_dir, f"h_pca_inverse_transform.png"), dpi=300, transparent=False
        )
        plt.close()

    if show:
        plt.show()
        plt.close()


def plot_results(
    bel,
    d: bool = True,
    h: bool = True,
    root: str = None,
    base_dir: str = None,
    folder: str = None,
    annotation: list = None,
):
    """
    Plots forecasts results in the 'uq' folder
    :param bel: BEL model
    :param annotation: List of annotations
    :param h: Boolean to plot target or not
    :param d: Boolean to plot predictor or not
    :param root: str: Forward ID
    :param folder: str: Well combination. '123456', '1'...
    :return:
    """
    if root is None:
        root = ""
    if folder is None:
        folder = ""
    # Directory
    md = jp(base_dir, root, folder)
    # Wells
    wells = Setup.Wells
    wells_id = list(wells.wells_data.keys())
    cols = [wells.wells_data[w]["color"] for w in wells_id if "pumping" not in w]

    if d:
        # Curves - d
        # Plot curves
        sdir = jp(md, "data")

        X = check_array(bel.X)
        try:
            X_obs = check_array(bel.X_obs)
        except ValueError:
            X_obs = check_array(bel.X_obs.to_numpy().reshape(1, -1))

        tc = X.reshape((Setup.HyperParameters.n_posts,) + bel.X_shape)
        tcp = X_obs.reshape((-1,) + bel.X_shape)
        tc = np.concatenate((tc, tcp), axis=0)

        # Plot parameters for predictor
        xlabel = "Observation index number"
        ylabel = "Concentration ($g/m^{3})$"
        factor = 1000
        labelsize = 11

        curves(
            cols=cols,
            tc=tc,
            sdir=sdir,
            xlabel=xlabel,
            ylabel=ylabel,
            factor=factor,
            labelsize=labelsize,
            highlight=[len(tc) - 1],
        )

        curves(
            cols=cols,
            tc=tc,
            sdir=sdir,
            xlabel=xlabel,
            ylabel=ylabel,
            factor=factor,
            labelsize=labelsize,
            highlight=[len(tc) - 1],
            ghost=True,
            title="curves_ghost",
        )

        curves_i(
            cols=cols,
            tc=tc,
            xlabel=xlabel,
            ylabel=ylabel,
            factor=factor,
            labelsize=labelsize,
            sdir=sdir,
            highlight=[len(tc) - 1],
        )

    if h:
        # WHP - h test + training
        fig_dir = jp(base_dir, root)
        ff = jp(fig_dir, "whpa.png")  # figure name
        Y = check_array(bel.Y)
        try:
            Y_obs = check_array(bel.Y_obs)
        except ValueError:
            Y_obs = check_array(bel.Y_obs.to_numpy().reshape(1, -1))
        h_test = Y_obs.reshape((bel.Y_shape[1], bel.Y_shape[2]))
        h_training = Y.reshape((-1,) + (bel.Y_shape[1], bel.Y_shape[2]))
        # Plots target training + prediction
        whpa_plot(whpa=h_training, color="blue", alpha=0.5)
        whpa_plot(
            whpa=h_test,
            color="r",
            lw=2,
            alpha=0.8,
            xlabel="X(m)",
            ylabel="Y(m)",
            labelsize=11,
        )
        colors = ["blue", "red"]
        labels = ["Training", "Test"]
        legend = _proxy_annotate(annotation=[], loc=2, fz=14)
        _proxy_legend(legend1=legend, colors=colors, labels=labels, fig_file=ff)
        plt.close()

        # WHPs
        ff = jp(md, "uq", f"{root}_cca_{bel.cca.n_components}.png")
        forecast_posterior = bel.random_sample(n_posts=Setup.HyperParameters.n_posts)
        forecast_posterior = bel.inverse_transform(forecast_posterior)
        forecast_posterior = forecast_posterior.reshape(
            (-1,) + (bel.Y_shape[1], bel.Y_shape[2])
        )

        # I display here the prior h behind the forecasts sampled from the posterior.
        well_ids = [0] + list(map(int, list(folder)))
        labels = ["Training", "Samples", "True test"]
        colors = ["darkblue", "darkred", "k"]

        # Training
        _, well_legend = whpa_plot(
            whpa=h_training,
            alpha=0.5,
            lw=0.5,
            color=colors[0],
            show_wells=True,
            well_ids=well_ids,
            show=False,
        )

        # Samples
        whpa_plot(
            whpa=forecast_posterior,
            color=colors[1],
            lw=1,
            alpha=0.8,
            highlight=True,
            show=False,
        )

        # True test
        whpa_plot(
            whpa=h_test,
            color=colors[2],
            lw=0.8,
            alpha=1,
            x_lim=[800, 1200],
            xlabel="X(m)",
            ylabel="Y(m)",
            labelsize=11,
            show=False,
        )

        # Other tricky operation to add annotation
        legend_an = _proxy_annotate(annotation=annotation, loc=2, fz=14)

        # Tricky operation to add a second legend:
        _proxy_legend(
            legend1=well_legend,
            extra=[legend_an],
            colors=colors,
            labels=labels,
            fig_file=ff,
        )


def mode_histo(
    colors: list,
    an_i: int,
    wm: np.array,
    title: str = None,
    fig_name: str = "average",
    directory: str = None,
):
    """

    :param colors:
    :param an_i: Figure annotation
    :param wm: Arrays of metric
    :param fig_name:
    :return:
    """
    if directory is None:
        directory = Setup.Directories.forecasts_dir
    alphabet = string.ascii_uppercase
    wid = list(map(str, Setup.Wells.combination))  # Wel identifiers (n)

    pipeline = Pipeline(
        [
            ("s_scaler", StandardScaler()),
        ]
    )
    wm = pipeline.fit_transform(wm)

    modes = []  # Get MHD corresponding to each well's mode
    for i, m in enumerate(wm):  # For each well, look up its MHD distribution
        count, values = np.histogram(m, bins="fd")
        # (Freedman Diaconis Estimator)
        # Robust (resilient to outliers) estimator that takes into account data variability and data size.
        # https://numpy.org/doc/stable/reference/generated/numpy.histogram_bin_edges.html#numpy.histogram_bin_edges
        idm = np.argmax(count)
        mode = values[idm]
        modes.append(mode)

    modes = np.array(modes)  # Scale modes
    modes -= np.mean(modes)

    # Bar plot
    plt.bar(np.arange(1, 7), -modes, color=colors)
    # plt.title("Amount of information of each well")
    plt.title(f"{fig_name}")
    plt.xlabel("Well ID")
    plt.ylabel("Opposite deviation from mode's mean")
    plt.grid(color="#95a5a6", linestyle="-", linewidth=0.5, axis="y", alpha=0.7)

    legend_a = _proxy_annotate(annotation=[alphabet[an_i + 1]], loc=2, fz=14)
    plt.gca().add_artist(legend_a)

    plt.savefig(
        os.path.join(directory, f"{fig_name}_well_mode.png"),
        dpi=300,
        transparent=False,
    )
    plt.close()

    # Plot BOX
    columns = ["1", "2", "3", "4", "5", "6"]
    wmd = pd.DataFrame(columns=columns, data=wm.T)
    palette = {columns[i]: colors[i] for i in range(len(columns))}
    # palette = {'b', 'g', 'r', 'c', 'm', 'y'}
    fig, ax1 = plt.subplots()
    sns.boxplot(data=wmd, palette=palette, order=columns, linewidth=1, ax=ax1)
    [line.set_color("white") for line in ax1.get_lines()[4::6]]
    plt.ylim([-2.5, 3])
    plt.xlabel("Well ID")
    plt.ylabel("Metric value")
    if title is None:
        title = "Box-plot of the metric values for each data source"
    plt.title(title)
    plt.grid(color="saddlebrown", linestyle="--", linewidth=0.7, axis="y", alpha=0.5)

    try:
        an_i = int(directory.split("split")[-1])
    except ValueError:
        pass
    legend_a = _proxy_annotate(annotation=[alphabet[an_i]], loc=2, fz=14)
    plt.gca().add_artist(legend_a)

    legend_b = _proxy_annotate(annotation=[f"Fold {an_i + 1}"], loc=1, fz=14)
    plt.gca().add_artist(legend_b)

    plt.savefig(
        os.path.join(directory, f"{fig_name}_well_box.png"),
        dpi=300,
        transparent=False,
    )
    plt.close()

    # Plot histogram
    for i, m in enumerate(wm):
        sns.kdeplot(m, color=f"{colors[i]}", shade=True, linewidth=2)
    # plt.title("Summed metric distribution for each well")
    plt.title(f"{fig_name}")
    plt.xlabel("Summed metric")
    plt.ylabel("KDE")
    legend_1 = plt.legend(wid, loc=2)
    plt.gca().add_artist(legend_1)
    plt.grid(alpha=0.2)

    legend_a = _proxy_annotate(annotation=[alphabet[an_i]], loc=2, fz=14)
    plt.gca().add_artist(legend_a)

    plt.savefig(
        os.path.join(directory, f"{fig_name}_hist.png"),
        dpi=300,
        transparent=False,
    )
    plt.close()


def curves(
    cols: list,
    tc: np.array,
    highlight: list = None,
    ghost: bool = False,
    sdir: str = None,
    labelsize: float = 12,
    factor: float = 1,
    conc: bool = 0,
    xlabel: str = None,
    ylabel: str = None,
    title: str = "curves",
    show: bool = False,
):
    """
    Shows every breakthrough curve stacked on a plot.
    :param cols: List of colors
    :param ylabel:
    :param xlabel:
    :param factor:
    :param labelsize:
    :param tc: Curves with shape (n_sim, n_wells, n_time_steps)
    :param highlight: list: List of indices of curves to highlight in the plot
    :param ghost: bool: Flag to only display highlighted curves.
    :param sdir: Directory in which to save figure
    :param title: str: Title
    :param show: Whether to show or not
    """
    if highlight is None:
        highlight = []

    if not conc:
        n_sim, n_wells, nts = tc.shape
        for i in range(n_sim):
            for t in range(n_wells):
                if i in highlight:
                    plt.plot(tc[i][t] * factor, color=cols[t], linewidth=2, alpha=1)
                elif not ghost:
                    plt.plot(tc[i][t] * factor, color=cols[t], linewidth=0.2, alpha=0.5)
    else:
        plt.plot(np.concatenate(tc[0]) * factor, color=cols[0], linewidth=2, alpha=1)

    plt.grid(linewidth=0.3, alpha=0.4)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.tick_params(labelsize=labelsize)
    if sdir:
        utils.dirmaker(sdir)
        plt.savefig(jp(sdir, f"{title}.png"), dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()
        plt.close()


def curves_i(
    cols: list,
    tc: np.array,
    highlight: list = None,
    labelsize: float = 12,
    factor: float = 1,
    xlabel: str = None,
    ylabel: str = None,
    sdir: str = None,
    show: bool = False,
):
    """
    Shows every breakthrough individually for each observation point.
    Will produce n_well figures of n_sim curves each.
    :param cols: List of colors
    :param labelsize:
    :param factor:
    :param xlabel:
    :param ylabel:
    :param tc: Curves with shape (n_sim, n_wells, n_time_steps)
    :param highlight: list: List of indices of curves to highlight in the plot
    :param sdir: Directory in which to save figure
    :param show: Whether to show or not

    """
    if highlight is None:
        highlight = []
    title = "curves"
    n_sim, n_wels, nts = tc.shape
    for t in range(n_wels):
        for i in range(n_sim):
            if i in highlight:
                plt.plot(tc[i][t] * factor, color="k", linewidth=2, alpha=1)
            else:
                plt.plot(tc[i][t] * factor, color=cols[t], linewidth=0.2, alpha=0.5)
        colors = [cols[t], "k"]
        plt.grid(linewidth=0.3, alpha=0.4)
        plt.tick_params(labelsize=labelsize)
        # plt.title(f'Well {t + 1}')

        alphabet = string.ascii_uppercase
        legend_a = _proxy_annotate([f"{alphabet[t]}. Well {t + 1}"], fz=12, loc=2)

        labels = ["Training", "Test"]
        _proxy_legend(legend1=legend_a, colors=colors, labels=labels, loc=1)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        if sdir:
            utils.dirmaker(sdir)
            plt.savefig(jp(sdir, f"{title}_{t + 1}.png"), dpi=300, transparent=False)
            plt.close()
        if show:
            plt.show()
            plt.close()


def plot_wells(wells: Setup.Wells, well_ids: list = None, markersize: float = 4.0):
    if well_ids is None:
        comb = [0] + list(wells.combination)
    else:
        comb = well_ids
    # comb = [0] + list(self.wells.combination)
    # comb = [0] + list(self.wells.combination)
    keys = [list(wells.wells_data.keys())[i] for i in comb]
    wbd = {k: wells.wells_data[k] for k in keys if k in wells.wells_data}
    s = 0
    for i in wbd:
        n = comb[s]
        if n == 0:
            label = "pw"
        else:
            label = f"{n}"
        if n in comb:
            plt.plot(
                wbd[i]["coordinates"][0],
                wbd[i]["coordinates"][1],
                f'{wbd[i]["color"]}o',
                markersize=markersize,
                markeredgecolor="k",
                markeredgewidth=0.5,
                label=label,
            )
        s += 1


def plot_pc_ba(
    bel,
    base_dir: str = None,
    root: str = None,
    w: str = None,
    data: bool = False,
    target: bool = False,
):
    """
    Comparison between original variables and the same variables back-transformed with n PCA components.
    :param root:
    :param data:
    :param target:
    :return:
    """

    if isinstance(root, (list, tuple)):
        if len(root) > 1:
            logger.error("Input error")
            return
        else:
            root = root[0]

    subdir = os.path.join(base_dir, root)

    if data:
        # Plot parameters for predictor
        xlabel = "Observation index number"
        ylabel = "Concentration ($g/m^{3})$"
        factor = 1000
        labelsize = 11
        d_pca_inverse_plot(
            bel,
            root=root,
            xlabel=xlabel,
            ylabel=ylabel,
            labelsize=labelsize,
            factor=factor,
            fig_dir=os.path.join(subdir, w, "pca"),
        )
    if target:
        h_pca_inverse_plot(bel, fig_dir=os.path.join(subdir, w, "pca"))


def plot_whpa(bel, base_dir):
    """
    Loads target pickle and plots all training WHPA
    :return:
    """

    h_training = bel.Y.reshape(bel.Y_shape)

    whpa_plot(
        whpa=h_training, highlight=True, halpha=0.5, lw=0.1, color="darkblue", alpha=0.5
    )

    h_pred = bel.Y_obs.reshape(bel.Y_shape)
    whpa_plot(
        whpa=h_pred,
        color="darkred",
        lw=1,
        alpha=1,
        annotation=[],
        xlabel="X(m)",
        ylabel="Y(m)",
        labelsize=11,
    )

    labels = ["Training", "Test"]
    legend = _proxy_annotate(annotation=[], loc=2, fz=14)
    _proxy_legend(
        legend1=legend,
        colors=["darkblue", "darkred"],
        labels=labels,
        fig_file=os.path.join(base_dir, "whpa_training.png"),
    )


def cca_vision(bel, base_dir: str = None, root: str = None):
    """
    Loads CCA pickles and plots components for all folders
    :param bel: BEL model
    :param base_dir
    :param root:
    :return:
    """

    if isinstance(root, (list, tuple)):
        if len(root) > 1:
            logger.error("Input error")
            return
        else:
            root = root[0]
    elif root is None:
        root = ""

    subdir = os.path.join(base_dir, root)

    res_dir = os.path.join(subdir, "obj")

    # CCA coefficient plot
    cca_coefficient = np.corrcoef(bel.X_c.T, bel.Y_c.T).diagonal(
        offset=bel.cca.n_components
    )  # Gets correlation coefficient
    plt.plot(cca_coefficient, "lightblue", zorder=1)
    plt.scatter(
        x=np.arange(len(cca_coefficient)),
        y=cca_coefficient,
        c=cca_coefficient,
        alpha=1,
        s=50,
        cmap="coolwarm",
        zorder=2,
    )
    cb = plt.colorbar()
    cb.ax.set_title(r"$\it{" + "r" + "}$")
    plt.grid(alpha=0.4, linewidth=0.5, zorder=0)
    plt.xticks(np.arange(len(cca_coefficient)), np.arange(1, len(cca_coefficient) + 1))
    plt.tick_params(labelsize=5)
    plt.yticks([])
    # plt.title('Decrease of CCA correlation coefficient with component number')
    plt.ylabel("Correlation coefficient")
    plt.xlabel("Component number")

    # Add annotation
    legend = _proxy_annotate(annotation=["D"], fz=14, loc=1)
    plt.gca().add_artist(legend)

    plt.savefig(
        os.path.join(os.path.dirname(res_dir), "cca", "coefs.png"),
        bbox_inches="tight",
        dpi=300,
        transparent=False,
    )
    plt.close()

    # KDE plots which consume a lot of time.
    _kde_cca(bel, sdir=os.path.join(subdir, "cca"))


def pca_vision(
    bel,
    base_dir: str,
    root: str or list = None,
    w: str = None,
    d: bool = True,
    h: bool = True,
    scores: bool = True,
    exvar: bool = True,
    before_after=True,
    labels: bool = True,
):
    """
    Loads PCA pickles and plot scores for all folders
    :param bel: BEL model
    :param w:
    :param labels:
    :param root: str:
    :param d: bool:
    :param h: bool:
    :param scores: bool:
    :param exvar: bool:
    :return:
    """

    if root is None:
        root = ""
    if w is None:
        w = ""

    subdir = jp(base_dir, root, w, "pca")

    if d:
        fig_file = os.path.join(subdir, "d_scores.png")
        if scores:
            pca_scores(
                training=bel.X_pc,
                prediction=bel.X_obs_pc,
                n_comp=Setup.HyperParameters.n_pc_predictor,
                # annotation=["C"],
                labels=labels,
                fig_file=fig_file,
            )
        # Explained variance plots
        if exvar:
            fig_file = os.path.join(subdir, "d_exvar.png")
            explained_variance(
                bel,
                n_comp=Setup.HyperParameters.n_pc_predictor,
                thr=0.8,
                # annotation=["E"],
                fig_file=fig_file,
            )
        if before_after:
            plot_pc_ba(bel, base_dir=base_dir, root=root, w=w, data=True, target=False)
    if h:
        # Transform and split
        h_pc_training = bel.Y_pc
        try:
            Y_obs = check_array(bel.Y_obs)
        except ValueError:
            Y_obs = check_array(bel.Y_obs.to_numpy().reshape(1, -1))
        h_pc_prediction = bel.Y_pre_processing.transform(Y_obs)
        # Plot
        fig_file = os.path.join(subdir, "h_pca_scores.png")
        if scores:
            pca_scores(
                training=h_pc_training,
                prediction=h_pc_prediction,
                n_comp=Setup.HyperParameters.n_pc_target,
                # annotation=["D"],
                labels=labels,
                fig_file=fig_file,
            )
        # Explained variance plots
        if exvar:
            fig_file = os.path.join(subdir, "h_pca_exvar.png")
            explained_variance(
                bel,
                n_comp=Setup.HyperParameters.n_pc_target,
                thr=0.8,
                # annotation=["F"],
                fig_file=fig_file,
            )
        if before_after:
            plot_pc_ba(bel, base_dir=base_dir, root=root, w=w, data=False, target=True)


def d_pca_inverse_plot(
    bel,
    root,
    factor: float = 1.0,
    xlabel: str = None,
    ylabel: str = None,
    labelsize: float = 11.0,
    fig_dir: str = None,
    show: bool = False,
):
    """
    Plot used to compare the reproduction of the original physical space after PCA transformation.
    :param bel: BEL model
    :param xlabel:
    :param ylabel:
    :param labelsize:
    :param factor:
    :param fig_dir: str:
    :param show: bool:
    :return:
    """

    shape = bel.X_shape
    v_pc = bel.X_obs_pc

    nc = bel.X_pre_processing["pca"].n_components_
    dummy = np.zeros((1, nc))
    dummy[:, : v_pc.shape[1]] = v_pc

    v_pred = bel.X_pre_processing.inverse_transform(dummy).reshape((-1,) + shape)
    to_plot = np.copy(bel.X_obs).reshape((-1,) + shape)

    cols = ["r" for _ in range(shape[1])]
    highlights = [i for i in range(shape[1])]
    curves(cols=cols, tc=to_plot, factor=factor, highlight=highlights, conc=True)

    cols = ["b" for _ in range(shape[1])]
    curves(cols=cols, tc=v_pred, factor=factor, highlight=highlights, conc=True)

    # Add title inside the box
    an = ["A"]
    legend_a = _proxy_annotate(annotation=an, loc=2, fz=14)
    _proxy_legend(
        legend1=legend_a,
        colors=["red", "blue"],
        labels=["Physical", "Back transformed"],
        marker=["-", "-"],
        loc=1,
    )
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.tick_params(labelsize=labelsize)

    # Increase y axis by a small percentage for annotation in upper left corner
    yrange = np.max(to_plot * factor) * 1.15
    plt.ylim([0, yrange])

    if fig_dir is not None:
        utils.dirmaker(fig_dir)
        plt.savefig(jp(fig_dir, f"{root}_d.png"), dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()
        plt.close()



"""Some visualization utilities"""

#  Copyright (c) 2021. Robin Thibaut, Ghent University
import itertools
import os
import string
from os.path import join as jp

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.pyplot import legend
from numpy import ma
from scipy.interpolate import BSpline, make_interp_spline

import skbel.utils
from sklearn.utils import check_array
from skbel.algorithms import KDE, kde_params, posterior_conditional

__all__ = [
    "_my_alphabet",
    "_proxy_legend",
    "_proxy_annotate",
    "explained_variance",
    "pca_scores",
    "cca_plot",
    "_despine",
    "_kde_cca",
]


def _my_alphabet(az: int):
    """
    Method used to make custom figure annotations.
    :param az: Index
    :return:
    """
    alphabet = string.ascii_uppercase
    extended_alphabet = ["".join(i) for i in list(itertools.permutations(alphabet, 2))]

    if az <= 25:
        sub = alphabet[az]
    else:
        j = az - 26
        sub = extended_alphabet[j]

    return sub


def _proxy_legend(
    legend1: legend = None,
    colors: list = None,
    labels: list = None,
    loc: int = 4,
    marker: list = None,
    pec: list = None,
    fz: float = 11,
    fig_file: str = None,
    extra: list = None,
    obj=None,
):
    """
    Add a second legend to a figure @ bottom right (loc=4)
    https://stackoverflow.com/questions/12761806/matplotlib-2-different-legends-on-same-graph
    :param legend1: First legend instance from the figure
    :param colors: List of colors
    :param labels: List of labels
    :param loc: Position of the legend
    :param marker: Points 'o' or line '-'
    :param pec: List of point edge color, e.g. [None, 'k']
    :param fz: Fontsize
    :param fig_file: Path to figure file
    :param extra: List of extra elements to be added on the final figure
    :return:
    """

    if obj is None:
        obj = plt
    # Default parameters
    if colors is None:
        colors = ["w"]
    if labels is None:
        labels = []
    if pec is None:
        pec = [None for _ in range(len(colors))]
    if extra is None:
        extra = []
    if marker is None:
        marker = ["-" for _ in range(len(colors))]

    # Proxy figures (empty plots)
    proxys = [
        plt.plot([], marker[i], color=c, markeredgecolor=pec[i])
        for i, c in enumerate(colors)
    ]

    leg = obj.legend([p[0] for p in proxys], labels, loc=loc, fontsize=fz)
    for text in leg.get_texts():
        text.set_color("k")

    if legend1:
        try:
            obj.gca().add_artist(legend1)
        except AttributeError:
            obj.add_artist(legend1)

    for el in extra:
        try:
            obj.gca().add_artist(el)
        except AttributeError:
            obj.add_artist(el)

    if fig_file:
        skbel.utils.dirmaker(os.path.dirname(fig_file))
        plt.savefig(fig_file, bbox_inches="tight", dpi=300, transparent=False)
        plt.close()


def _proxy_annotate(annotation: list = None, loc: int = 1, fz: float = 11, obj=None):
    """
    Places annotation (or title) within the figure box
    :param annotation: Must be a list of labels even of it only contains one label. Savvy ?
    :param fz: Fontsize
    :param loc: Location (default: 1 = upper right corner, 2 = upper left corner)
    :return:
    """
    if obj is None:
        obj = plt
    if annotation is None:
        annotation = []
    legend_a = obj.legend(
        plt.plot([], linestyle=None, color="w", alpha=0, markeredgecolor=None),
        annotation,
        handlelength=0,
        handletextpad=0,
        fancybox=True,
        loc=loc,
        fontsize=fz,
    )

    for text in legend_a.get_texts():
        text.set_color("k")

    return legend_a


def explained_variance(
    bel,
    n_comp: int = 0,
    thr: float = 1.0,
    annotation: list = None,
    fig_file: str = None,
    show: bool = False,
):
    """
    PCA explained variance plot
    :param bel
    :param n_comp: Number of components to display
    :param thr: float: Threshold
    :param annotation: List of annotation(s)
    :param fig_file:
    :param show:
    :return:
    """
    plt.grid(alpha=0.1)
    if not n_comp:
        n_comp = bel.X_pre_processing["pca"].n_components_

    # Index where explained variance is below threshold:
    ny = len(
        np.where(
            np.cumsum(bel.X_pre_processing["pca"].explained_variance_ratio_) < thr
        )[0]
    )
    # Explained variance vector:
    cum = (
        np.cumsum(bel.X_pre_processing["pca"].explained_variance_ratio_[:n_comp]) * 100
    )
    # x-ticks
    plt.xticks(
        np.concatenate([np.array([0]), np.arange(4, n_comp, 5)]),
        np.concatenate([np.array([1]), np.arange(5, n_comp + 5, 5)]),
        fontsize=11,
    )
    # Tricky y-ticks
    yticks = np.append(cum[:ny], cum[-1])
    plt.yticks(yticks, fontsize=8.5)
    # bars for aesthetics
    plt.bar(
        np.arange(n_comp),
        np.cumsum(bel.X_pre_processing["pca"].explained_variance_ratio_[:n_comp]) * 100,
        color="m",
        alpha=0.1,
    )
    # line for aesthetics
    plt.plot(
        np.arange(n_comp),
        np.cumsum(bel.X_pre_processing["pca"].explained_variance_ratio_[:n_comp]) * 100,
        "-o",
        linewidth=0.5,
        markersize=1.5,
        alpha=0.8,
    )
    # Axes labels
    plt.xlabel("PC number", fontsize=10)
    plt.ylabel("Cumulative explained variance (%)", fontsize=11)
    # Legend
    legend_a = _proxy_annotate(annotation=annotation, loc=2, fz=14)
    plt.gca().add_artist(legend_a)

    if fig_file:
        skbel.utils.dirmaker(os.path.dirname(fig_file))
        plt.savefig(fig_file, dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()
        plt.close()


def pca_scores(
    training: np.array,
    prediction: np.array,
    n_comp: int,
    annotation: list,
    fig_file: str = None,
    labels: bool = True,
    show: bool = False,
):
    """
    PCA scores plot, displays scores of observations above those of training.
    :param labels:
    :param training: Training scores
    :param prediction: Test scores
    :param n_comp: How many components to show
    :param annotation: List of annotation(s)
    :param fig_file:
    :param show:
    :return:
    """
    # Scores plot
    # Grid
    plt.grid(alpha=0.2)
    # Ticks
    plt.xticks(
        np.concatenate([np.array([0]), np.arange(4, n_comp, 5)]),
        np.concatenate([np.array([1]), np.arange(5, n_comp + 5, 5)]),
    )  # Plot all training scores
    plt.plot(training.T[:n_comp], "ob", markersize=3, alpha=0.1)
    # plt.plot(training.T[:ut], '+w', markersize=.5, alpha=0.2)

    # For each sample used for prediction:
    for sample_n in range(len(prediction)):
        # Select observation
        pc_obs = prediction[sample_n]
        # Create beautiful spline to follow prediction scores
        xnew = np.linspace(1, n_comp, 200)  # New points for plotting curve
        spl = make_interp_spline(
            np.arange(1, n_comp + 1), pc_obs.T[:n_comp], k=3
        )  # type: BSpline
        power_smooth = spl(xnew)
        # I forgot why I had to put '-1'
        plt.plot(xnew - 1, power_smooth, "red", linewidth=1.2, alpha=0.9)

        plt.plot(
            pc_obs.T[:n_comp],  # Plot observations scores
            "ro",
            markersize=3,
            markeredgecolor="k",
            markeredgewidth=0.4,
            alpha=0.8,
            label=str(sample_n),
        )

    if labels:
        plt.title("Principal Components of training and test dataset")
        plt.xlabel("PC number")
        plt.ylabel("PC")
    plt.tick_params(labelsize=11)
    # Add legend
    # Add title inside the box
    legend_a = _proxy_annotate(annotation=annotation, loc=2, fz=14)
    _proxy_legend(
        legend1=legend_a,
        colors=["blue", "red"],
        labels=["Training", "Test"],
        marker=["o", "o"],
    )

    if fig_file:
        skbel.utils.dirmaker(os.path.dirname(fig_file))
        plt.savefig(fig_file, dpi=300, transparent=False)
        plt.close()
    if show:
        plt.show()


def cca_plot(
    bel,
    d: np.array,
    h: np.array,
    d_pc_prediction: np.array,
    sdir: str = None,
    show: bool = False,
):
    """
    CCA plots.
    Receives d, h PC components to be predicted, transforms them in CCA space and adds it to the plots.
    :param bel
    :param d: d CCA scores
    :param h: h CCA scores
    :param d_pc_prediction: d test PC scores
    :param sdir: str:
    :param show: bool:
    :return:
    """

    cca_coefficient = np.corrcoef(d, h).diagonal(
        offset=bel.cca.n_components
    )  # Gets correlation coefficient

    # CCA plots for each observation:
    for i in range(bel.cca.n_components):
        for sample_n in range(len(d_pc_prediction)):  # For each 'observation'
            pass

        subtitle = _my_alphabet(i)

        # Add title inside the box
        an = [f"{subtitle}. Pair {i + 1} - R = {round(cca_coefficient[i], 3)}"]
        legend_a = _proxy_annotate(annotation=an, loc=2, fz=14)

        _proxy_legend(
            legend1=legend_a,
            colors=["black", "white"],
            labels=["Training", "Test"],
            marker=["o", "o"],
            pec=["k", "k"],
        )

        if sdir:
            skbel.utils.dirmaker(sdir)
            plt.savefig(
                jp(sdir, "cca{}.png".format(i)),
                bbox_inches="tight",
                dpi=300,
                transparent=False,
            )
            plt.close()
        if show:
            plt.show()
            plt.close()


def _despine(
    fig=None,
    ax=None,
    top=True,
    right=True,
    left=False,
    bottom=False,
    offset=None,
    trim=False,
):
    """Remove the top and right spines from plot(s).

    fig : matplotlib figure, optional
        Figure to despine all axes of, defaults to the current figure.
    ax : matplotlib axes, optional
        Specific axes object to despine. Ignored if fig is provided.
    top, right, left, bottom : boolean, optional
        If True, remove that spine.
    offset : int or dict, optional
        Absolute distance, in points, spines should be moved away
        from the axes (negative values move spines inward). A single value
        applies to all spines; a dict can be used to set offset values per
        side.
    trim : bool, optional
        If True, limit spines to the smallest and largest major tick
        on each non-despined axis.

    Returns
    -------
    None

    """
    # Get references to the axes we want
    if fig is None and ax is None:
        axes_dummy = plt.gcf().axes
    elif fig is not None:
        axes_dummy = fig.axes
    elif ax is not None:
        axes_dummy = [ax]
    else:
        return

    for ax_i in axes_dummy:
        for side in ["top", "right", "left", "bottom"]:
            # Toggle the spine objects
            is_visible = not locals()[side]
            ax_i.spines[side].set_visible(is_visible)
            if offset is not None and is_visible:
                try:
                    val = offset.get(side, 0)
                except AttributeError:
                    val = offset
                ax_i.spines[side].set_position(("outward", val))

        # Potentially move the ticks
        if left and not right:
            maj_on = any(t.tick1line.get_visible() for t in ax_i.yaxis.majorTicks)
            min_on = any(t.tick1line.get_visible() for t in ax_i.yaxis.minorTicks)
            ax_i.yaxis.set_ticks_position("right")
            for t in ax_i.yaxis.majorTicks:
                t.tick2line.set_visible(maj_on)
            for t in ax_i.yaxis.minorTicks:
                t.tick2line.set_visible(min_on)

        if bottom and not top:
            maj_on = any(t.tick1line.get_visible() for t in ax_i.xaxis.majorTicks)
            min_on = any(t.tick1line.get_visible() for t in ax_i.xaxis.minorTicks)
            ax_i.xaxis.set_ticks_position("top")
            for t in ax_i.xaxis.majorTicks:
                t.tick2line.set_visible(maj_on)
            for t in ax_i.xaxis.minorTicks:
                t.tick2line.set_visible(min_on)

        if trim:
            # clip off the parts of the spines that extend past major ticks
            xticks = np.asarray(ax_i.get_xticks())
            if xticks.size:
                firsttick = np.compress(xticks >= min(ax_i.get_xlim()), xticks)[0]
                lasttick = np.compress(xticks <= max(ax_i.get_xlim()), xticks)[-1]
                ax_i.spines["bottom"].set_bounds(firsttick, lasttick)
                ax_i.spines["top"].set_bounds(firsttick, lasttick)
                newticks = xticks.compress(xticks <= lasttick)
                newticks = newticks.compress(newticks >= firsttick)
                ax_i.set_xticks(newticks)

            yticks = np.asarray(ax_i.get_yticks())
            if yticks.size:
                firsttick = np.compress(yticks >= min(ax_i.get_ylim()), yticks)[0]
                lasttick = np.compress(yticks <= max(ax_i.get_ylim()), yticks)[-1]
                ax_i.spines["left"].set_bounds(firsttick, lasttick)
                ax_i.spines["right"].set_bounds(firsttick, lasttick)
                newticks = yticks.compress(yticks <= lasttick)
                newticks = newticks.compress(newticks >= firsttick)
                ax_i.set_yticks(newticks)


def _get_defaults_kde_plot():
    height = 6
    ratio = 6
    space = 0

    xlim = None
    ylim = None
    marginal_ticks = False

    # Set up the subplot grid
    f = plt.figure(figsize=(height, height))
    gs = plt.GridSpec(ratio + 1, ratio + 1)

    # ax_joint = f.add_subplot(gs[1:, :-1])
    # ax_marg_x = f.add_subplot(gs[0, :-1], sharex=ax_joint)
    # ax_marg_y = f.add_subplot(gs[1:, -1], sharey=ax_joint)

    ax_joint = f.add_subplot(gs[1:, 1:-1])
    ax_marg_x = f.add_subplot(gs[0, 1:-1], sharex=ax_joint)
    ax_marg_y = f.add_subplot(gs[1:, -1], sharey=ax_joint)
    ax_cb = f.add_subplot(gs[1:, 0])

    fig = f
    ax_joint = ax_joint
    ax_marg_x = ax_marg_x
    ax_marg_y = ax_marg_y
    ax_cb = ax_cb

    # Turn off tick visibility for the measure axis on the marginal plots
    plt.setp(ax_marg_x.get_xticklabels(), visible=False)
    plt.setp(ax_marg_y.get_yticklabels(), visible=False)

    plt.setp(ax_marg_x.get_xticklabels(minor=True), visible=False)
    plt.setp(ax_marg_y.get_yticklabels(minor=True), visible=False)

    # Turn off the ticks on the density axis for the marginal plots
    plt.setp(ax_marg_x.yaxis.get_majorticklines(), visible=False)
    plt.setp(ax_marg_x.yaxis.get_minorticklines(), visible=False)

    plt.setp(ax_marg_y.xaxis.get_majorticklines(), visible=False)
    plt.setp(ax_marg_y.xaxis.get_minorticklines(), visible=False)

    plt.setp(ax_marg_x.get_yticklabels(), visible=False)
    plt.setp(ax_marg_y.get_xticklabels(), visible=False)

    plt.setp(ax_marg_x.get_yticklabels(minor=True), visible=False)
    plt.setp(ax_marg_y.get_xticklabels(minor=True), visible=False)

    ax_marg_x.yaxis.grid(False)
    ax_marg_y.xaxis.grid(False)

    ax_cb.axis("off")

    if xlim is not None:
        ax_joint.set_xlim(xlim)
    if ylim is not None:
        ax_joint.set_ylim(ylim)

    # Make the grid look nice
    _despine(f)
    if not marginal_ticks:
        _despine(ax=ax_marg_x, left=True)
        _despine(ax=ax_marg_y, bottom=True)

    for axes in [ax_marg_x, ax_marg_y, ax_cb]:
        for axis in [axes.xaxis, axes.yaxis]:
            axis.label.set_visible(False)
    f.tight_layout()
    f.subplots_adjust(hspace=space, wspace=space)

    return ax_joint, ax_marg_x, ax_marg_y, ax_cb


def _kde_cca(
    bel,
    sdir: str = None,
    show: bool = False,
    dist_plot: bool = False,
):
    # Find max kde value (absolutely not optimal)
    vmax = 0
    for comp_n in range(bel.cca.n_components):
        # Working with final product of BEL (not with raw cca scores)
        hp, sup = posterior_conditional(
            X=bel.X_f.T[comp_n], Y=bel.Y_f.T[comp_n], X_obs=bel.X_obs_f.T[comp_n]
        )

        # Plot h posterior given d
        density, _ = kde_params(x=bel.X_f.T[comp_n], y=bel.Y_f.T[comp_n])
        maxloc = np.max(density)
        if vmax < maxloc:
            vmax = maxloc

    cca_coefficient = np.corrcoef(bel.X_f.T, bel.Y_f.T).diagonal(
        offset=bel.cca.n_components
    )  # Gets correlation coefficient

    try:
        Y_obs = check_array(bel.Y_obs)
    except ValueError:
        Y_obs = check_array(bel.Y_obs.to_numpy().reshape(1, -1))

    # Transform Y obs
    bel.Y_obs_f = bel.transform(Y=Y_obs)

    # load prediction object
    post_test = bel.random_sample(n_posts=bel._n_posts)

    for comp_n in range(bel.cca.n_components):
        # Get figure default parameters
        ax_joint, ax_marg_x, ax_marg_y, ax_cb = _get_defaults_kde_plot()

        # Conditional:
        hp, sup = posterior_conditional(
            X=bel.X_f.T[comp_n], Y=bel.Y_f.T[comp_n], X_obs=bel.X_obs_f.T[comp_n]
        )

        # Plot h posterior given d
        density, support = kde_params(
            x=bel.X_f.T[comp_n], y=bel.Y_f.T[comp_n], gridsize=1000
        )
        xx, yy = support

        marginal_eval_x = KDE()
        marginal_eval_y = KDE()

        # support is cached
        kde_x, sup_x = marginal_eval_x(bel.X_f.T[comp_n])
        kde_y, sup_y = marginal_eval_y(bel.Y_f.T[comp_n])

        y_samp = post_test.T[comp_n]
        # use the same support as y
        kde_y_samp, sup_samp = marginal_eval_y(y_samp)

        # Filled contour plot
        # Mask values under threshold
        z = ma.masked_where(density <= np.finfo(np.float16).eps, density)
        # Filled contour plot
        # 'BuPu_r' is nice
        cf = ax_joint.contourf(
            xx, yy, z, cmap="coolwarm", levels=100, vmin=0, vmax=vmax
        )
        cb = plt.colorbar(cf, ax=[ax_cb], location="left")
        cb.ax.set_title("$KDE_{Gaussian}$", fontsize=10)
        # Vertical line
        ax_joint.axvline(
            x=bel.X_obs_f.T[comp_n],
            color="red",
            linewidth=1,
            alpha=0.5,
            label="$d^{c}_{True}$",
        )
        # Horizontal line
        ax_joint.axhline(
            y=bel.Y_obs_f.T[comp_n],
            color="deepskyblue",
            linewidth=1,
            alpha=0.5,
            label="$h^{c}_{True}$",
        )
        # Scatter plot
        ax_joint.plot(
            bel.X_f.T[comp_n],
            bel.Y_f.T[comp_n],
            "ko",
            markersize=2,
            markeredgecolor="w",
            markeredgewidth=0.2,
            alpha=0.9,
        )
        # Point
        ax_joint.plot(
            bel.X_obs_f.T[comp_n],
            bel.Y_obs_f.T[comp_n],
            "wo",
            markersize=5,
            markeredgecolor="k",
            alpha=1,
        )
        # Marginal x plot
        #  - Line plot
        ax_marg_x.plot(sup_x, kde_x, color="black", linewidth=0.5, alpha=1)
        #  - Fill to axis
        ax_marg_x.fill_between(
            sup_x, 0, kde_x, color="royalblue", alpha=0.5, label="$p(d^{c})$"
        )
        #  - Notch indicating true value
        ax_marg_x.axvline(
            x=bel.X_obs_f.T[comp_n], ymax=0.25, color="red", linewidth=1, alpha=0.5
        )
        ax_marg_x.legend(loc=2, fontsize=10)

        # Marginal y plot
        #  - Line plot
        ax_marg_y.plot(kde_y, sup_y, color="black", linewidth=0.5, alpha=1)
        #  - Fill to axis
        ax_marg_y.fill_betweenx(
            sup_y, 0, kde_y, alpha=0.5, color="darkred", label="$p(h^{c})$"
        )
        #  - Notch indicating true value
        ax_marg_y.axhline(
            y=bel.Y_obs_f.T[comp_n],
            xmax=0.25,
            color="deepskyblue",
            linewidth=1,
            alpha=0.5,
        )
        # Marginal y plot with BEL
        #  - Line plot
        ax_marg_y.plot(kde_y_samp, sup_samp, color="black", linewidth=0.5, alpha=0)
        #  - Fill to axis
        ax_marg_y.fill_betweenx(
            sup_samp,
            0,
            kde_y_samp,
            color="teal",
            alpha=0.5,
            label="$p(h^{c}|d^{c}_{*})_{BEL}$",
        )
        # Conditional distribution
        #  - Line plot
        ax_marg_y.plot(hp, sup, color="red", alpha=0)
        #  - Fill to axis
        ax_marg_y.fill_betweenx(
            sup,
            0,
            hp,
            color="mediumorchid",
            alpha=0.5,
            label="$p(h^{c}|d^{c}_{*})_{KDE}$",
        )
        ax_marg_y.legend(fontsize=10)
        # Labels
        ax_joint.set_xlabel("$d^{c}$", fontsize=14)
        ax_joint.set_ylabel("$h^{c}$", fontsize=14)
        plt.tick_params(labelsize=14)

        # Add custom artists
        subtitle = _my_alphabet(comp_n)
        # Add title inside the box
        an = [
            f"{subtitle}. Pair {comp_n + 1} - "
            + r"$\it{"
            + "r"
            + "}$"
            + f" = {round(cca_coefficient[comp_n], 3)}"
        ]
        legend_a = _proxy_annotate(obj=ax_joint, annotation=an, loc=2, fz=12)
        #
        _proxy_legend(
            obj=ax_joint,
            legend1=legend_a,
            colors=["black", "white", "red", "deepskyblue"],
            labels=["$Training$", "$Test$", "$d^{c}_{*}$", "$h^{c}_{True}$"],
            marker=["o", "o", "-", "-"],
            pec=["k", "k", None, None],
            fz=10,
        )

        if sdir:
            skbel.utils.dirmaker(sdir)
            plt.savefig(
                jp(sdir, f"cca_kde_{comp_n}.png"),
                bbox_inches="tight",
                dpi=300,
                transparent=False,
            )
            plt.close()
        if show:
            plt.show()
            plt.close()

        def posterior_distribution():
            # prior
            plt.plot(sup_y, kde_y, color="black", linewidth=0.5, alpha=1)
            plt.fill_between(
                sup_y, 0, kde_y, color="mistyrose", alpha=1, label="$p(h^{c})$"
            )
            # posterior kde
            plt.plot(sup, hp, color="darkred", linewidth=0.5, alpha=0)
            plt.fill_between(
                sup,
                0,
                hp,
                color="salmon",
                alpha=0.5,
                label="$p(h^{c}|d^{c}_{*})$ (KDE)",
            )
            # posterior bel
            plt.plot(sup_samp, kde_y_samp, color="black", linewidth=0.5, alpha=1)
            plt.fill_between(
                sup_samp,
                0,
                kde_y_samp,
                color="gray",
                alpha=0.5,
                label="$p(h^{c}|d^{c}_{*})$ (BEL)",
            )

            # True prediction
            plt.axvline(
                x=bel.Y_obs_f[0],
                linewidth=3,
                alpha=0.4,
                color="deepskyblue",
                label="$h^{c}_{True}$",
            )

            # Grid
            plt.grid(alpha=0.2)

            # Tuning
            plt.ylabel("Density", fontsize=14)
            plt.xlabel("$h^{c}$", fontsize=14)
            plt.xlim([np.min(bel.X_f.T[comp_n]), np.max(bel.X_f.T[comp_n])])
            plt.tick_params(labelsize=14)

            plt.legend(loc=2)

            if sdir:
                skbel.utils.dirmaker(sdir)
                plt.savefig(
                    jp(sdir, f"cca_prior_post_{comp_n}.png"),
                    bbox_inches="tight",
                    dpi=300,
                    transparent=False,
                )
                plt.close()
            if show:
                plt.show()
                plt.close()

        if dist_plot:
            posterior_distribution()



#  Copyright (c) 2021. Robin Thibaut, Ghent University
"""
Boilerplate
-----------
Grid geometry parameters need to be passed around modules.
Defining data classes allows to avoid declaring those parameters several times.
"""

import os
import platform
from dataclasses import dataclass
from os.path import dirname, join

import numpy as np


class Machine(object):
    computer: str = platform.node()


class Setup:
    @dataclass
    class Directories:
        """Define main directories and file names"""

        # Content directory
        main_dir: str = dirname(os.path.abspath(__file__))
        storage_dir: str = join(main_dir, "storage")
        data_dir: str = join(main_dir, "datasets")
        hydro_res_dir: str = join(data_dir, "forwards")
        forecasts_dir: str = join(storage_dir, "forecasts")
        forecasts_base_dir: str = join(forecasts_dir, "base")
        grid_dir: str = join(main_dir, "spatial", "parameters")
        test_dir: str = join(main_dir, "testing")
        ref_dir: str = join(test_dir, "reference")

    @dataclass
    class Files:
        """Class to keep track of important file names"""

        # Output file names
        project_name: str = "whpa"
        predictor_name: str = "d"
        target_name: str = "h"

        # Files stored in the raw, forward dataset folder
        hk_file: str = "hk0.npy"
        predictor_file: str = "bkt.npy"
        target_file: str = "pz.npy"

        output_files = [hk_file, predictor_file, target_file]

        sgems_file: str = "hd.sgems"
        command_file: str = "sgsim_commands.py"

        sgems_family = [sgems_file, command_file, hk_file]

        # Files that are stored after preprocessing
        predictor_pickle: str = "d_pca.pkl"
        target_pickle: str = "h_pca.pkl"
        model_pickle: str = "cca.pkl"
        target_post_bin: str = "post.npy"
        target_pc_bin: str = "target_pc.npy"
        predictor_training_bin: str = "training_curves.npy"
        predictor_test_bin: str = "test_curves.npy"

    @dataclass
    class GridDimensions:
        """Class for keeping track of grid dimensions"""

        x_lim: float = 1500.0
        y_lim: float = 1000.0
        z_lim: float = 1.0

        dx: float = 10.0  # Block x-dimension
        dy: float = 10.0  # Block y-dimension
        dz: float = 10.0  # Block z-dimension

        xo: float = 0.0
        yo: float = 0.0
        zo: float = 0.0

        nrow: int = y_lim // dy  # Number of rows
        ncol: int = x_lim // dx  # Number of columns
        nlay: int = 1  # Number of layers

        # Refinement parameters around the pumping well.
        # 150 meters from the pumping well coordinates, grid cells will have dimensions 9x9 and so on...
        r_params = np.array(
            [
                [9, 150],
                [8, 100],
                [7, 90],
                [6, 80],
                [5, 70],
                [4, 60],
                [3, 50],
                [2.5, 40],
                [2, 30],
                [1.5, 20],
                [1, 10],
            ]
        )  # ...10 meters from the pumping well coordinates, grid cells will have dimensions 1x1.

    @dataclass
    class Focus:
        """Geometry of the focused area on the main grid, enclosing all wells, as to reduce computation time"""

        x_range = [800, 1150]
        y_range = [300, 700]
        # Defines cell dimensions for the signed distance computation.
        cell_dim: float = 4

    @dataclass
    class Wells:
        """Wells coordinates"""

        wells_data = {
            "pumping0": {
                "coordinates": [1000, 500],
                "rates": [-1000, -1000, -1000],
                "color": "w",
            },
            "injection0": {
                "coordinates": [950, 450],
                "rates": [0, 24, 0],
                "color": "b",
            },
            "injection1": {
                "coordinates": [930, 560],
                "rates": [0, 24, 0],
                "color": "g",
            },
            "injection2": {
                "coordinates": [900, 505],
                "rates": [0, 24, 0],
                "color": "r",
            },
            "injection3": {
                "coordinates": [1068, 515],
                "rates": [0, 24, 0],
                "color": "c",
            },
            "injection4": {
                "coordinates": [1030, 580],
                "rates": [0, 24, 0],
                "color": "m",
            },
            "injection5": {
                "coordinates": [1050, 470],
                "rates": [0, 24, 0],
                "color": "y",
            },
        }

        # Injection wells in use for prediction (default: all)
        combination = np.arange(1, len(wells_data))
        colors = ["b", "g", "r", "c", "m", "y"]

    @dataclass
    class HyperParameters:
        """Learning hyper parameters"""

        # Predictor
        n_pc_predictor: int = 50
        n_tstp: int = 200

        # Target
        n_pc_target: int = 30

        # Posterior
        # Size of data set
        n_total: int = 500  # Parameter to optimize
        n_training: int = int(n_total * 0.8)
        n_test: int = int(n_total * 0.2)
        # Sample size
        n_posts: int = n_training

    @dataclass
    class ModelParameters:
        """Model hyper parameters"""

        # Prior K
        k_min: float = 1.4
        k_max: float = 2
        k_std: float = 0.4

    @dataclass
    class ED:
        """Experimental Design"""

        metric = None



#  Copyright (c) 2021. Robin Thibaut, Ghent University

import itertools
import os
import shutil
import types
from typing import List

import numpy as np
from loguru import logger

FLOAT_DTYPES = (np.float64, np.float32, np.float16)
Combination = List[List[int]]
Function = types.FunctionType


def flatten_array(arr: np.array) -> np.array:
    arr_flat = np.array([item for sublist in arr for item in sublist])
    return arr_flat.reshape(1, -1)


def data_read(file: str = None, start: int = 0, end: int = None):
    # end must be set to None and NOT -1
    """Reads space-separated dat file"""
    with open(file, "r") as fr:
        lines = np.copy(fr.readlines())[start:end]
        try:
            op = np.array(
                [list(map(float, line.split())) for line in lines], dtype=object
            )
        except ValueError:
            op = [line.split() for line in lines]
    return op


def folder_reset(folder: str, exceptions: list = None):
    """Deletes files in folder"""
    if not isinstance(exceptions, (list, tuple)):
        exceptions = [exceptions]
    try:
        for filename in os.listdir(folder):
            if filename not in exceptions:
                file_path = os.path.join(folder, filename)
                try:
                    if os.path.isfile(file_path):
                        os.unlink(file_path)
                except Exception as e:
                    logger.warning("Failed to delete %s. Reason: %s" % (file_path, e))
    except FileNotFoundError:
        pass


def dirmaker(dird: str, erase: bool = False):
    """
    Given a folder path, check if it exists, and if not, creates it.
    :param dird: str: Directory path.
    :param erase: bool: Whether to delete existing folder or not.
    :return:
    """
    try:
        if not os.path.exists(dird):
            os.makedirs(dird)
            return 0
        else:
            if erase:
                shutil.rmtree(dird)
                os.makedirs(dird)
            return 1
    except Exception as e:
        logger.warning(e)
        return 0


def combinator(combi):
    """Given a n-sized 1D array, generates all possible configurations, from size 1 to n-1.
    'None' will indicate to use the original combination.
    """
    cb = [
        list(itertools.combinations(combi, i)) for i in range(1, combi[-1] + 1)
    ]  # Get all possible wel combinations
    # Flatten and reverse to get all combination at index 0.
    cb = [item for sublist in cb for item in sublist][::-1]
    return cb



#  Copyright (c) 2021. Robin Thibaut, Ghent University

import os
from os.path import join as jp

import joblib
import pandas as pd
from loguru import logger
from sklearn.cross_decomposition import CCA
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, PowerTransformer

import demo_visualization as myvis
from demo_config import Setup
from skbel import utils
from skbel.learning.bel import BEL


def init_bel():
    """
    Set all BEL pipelines
    """
    # Pipeline before CCA
    X_pre_processing = Pipeline(
        [
            ("scaler", StandardScaler(with_mean=False)),
            ("pca", PCA()),
        ]
    )
    Y_pre_processing = Pipeline(
        [
            ("scaler", StandardScaler(with_mean=False)),
            ("pca", PCA()),
        ]
    )

    # Canonical Correlation Analysis
    # Number of CCA components is chosen as the min number of PC
    n_pc_pred, n_pc_targ = 50, 30

    cca = CCA(n_components=min(n_pc_targ, n_pc_pred), max_iter=500 * 20, tol=1e-6)

    # Pipeline after CCA
    X_post_processing = Pipeline(
        [("normalizer", PowerTransformer(method="yeo-johnson", standardize=True))]
    )
    Y_post_processing = Pipeline(
        [("normalizer", PowerTransformer(method="yeo-johnson", standardize=True))]
    )

    # Initiate BEL object
    bel_model = BEL(
        X_pre_processing=X_pre_processing,
        X_post_processing=X_post_processing,
        Y_pre_processing=Y_pre_processing,
        Y_post_processing=Y_post_processing,
        cca=cca,
        mode="kde",
    )

    # Set PC cut
    bel_model.X_n_pc = n_pc_pred
    bel_model.Y_n_pc = n_pc_targ

    return bel_model


def bel_training(
    bel_,
    *,
    X_train_: pd.DataFrame,
    x_test_: pd.DataFrame,
    y_train_: pd.DataFrame,
    y_test_: pd.DataFrame = None,
    directory: str = None,
):
    """
    :param bel_: BEL model
    :param X_train_: Predictor set for training
    :param x_test_: Predictor "test"
    :param y_train_: Target set for training
    :param y_test_: "True" target (optional)
    :param directory: Path to the directory in which to unload the results
    :return:
    """
    #%% Directory in which to load forecasts
    if directory is None:
        sub_dir = os.getcwd()
    else:
        sub_dir = directory

    # Folders
    obj_dir = jp(sub_dir, "obj")  # Location to save the BEL model
    fig_data_dir = jp(sub_dir, "data")  # Location to save the raw data figures
    fig_pca_dir = jp(sub_dir, "pca")  # Location to save the PCA figures
    fig_cca_dir = jp(sub_dir, "cca")  # Location to save the CCA figures
    fig_pred_dir = jp(sub_dir, "uq")  # Location to save the prediction figures

    # Creates directories
    [
        utils.dirmaker(f, erase=True)
        for f in [
            obj_dir,
            fig_data_dir,
            fig_pca_dir,
            fig_cca_dir,
            fig_pred_dir,
        ]
    ]

    # %% Fit BEL model
    bel_.Y_obs = y_test_
    bel_.fit(X=X_train_, Y=y_train_)

    # %% Sample for the observation
    # Extract n random sample (target pc's).
    # The posterior distribution is computed within the method below.
    bel_.predict(x_test_)

    # Save the fitted BEL model
    joblib.dump(bel_, jp(obj_dir, "bel.pkl"))
    msg = f"model trained and saved in {obj_dir}"
    logger.info(msg)


if __name__ == "__main__":
    # Initiate BEL model
    model = init_bel()

    # Set directories
    data_dir = jp(os.getcwd(), "dataset")
    output_dir = jp(os.getcwd(), "results")

    # Load dataset
    X_train = pd.read_pickle(jp(data_dir, "X_train.pkl"))
    X_test = pd.read_pickle(jp(data_dir, "X_test.pkl"))
    y_train = pd.read_pickle(jp(data_dir, "y_train.pkl"))
    y_test = pd.read_pickle(jp(data_dir, "y_test.pkl"))

    # Train model
    bel_training(
        bel_=model,
        X_train_=X_train,
        x_test_=X_test,
        y_train_=y_train,
        y_test_=y_test,
        directory=output_dir,
    )

    # Plot the results
    bel = joblib.load(jp(output_dir, "obj", "bel.pkl"))
    bel.n_posts = Setup.HyperParameters.n_posts

    # Plot raw data
    myvis.plot_results(bel, base_dir=output_dir)

    # Plot PCA
    # myvis.pca_vision(
    #     bel,
    #     base_dir=output_dir,
    # )

    # Plot CCA
    myvis.cca_vision(bel=bel, base_dir=output_dir)



import numpy as np

__all__ = [
    "get_block",
    "matrix_paste",
]

from scipy.spatial import distance_matrix


def get_block(pm: np.array, i: int) -> np.array:
    """
    Extracts block from a 2x2 partitioned matrix.
    :param pm: Partitioned matrix
    :param i: Block index
    1 2
    3 4
    :return: Bock #b
    """

    b = pm.shape[0] // 2

    if i == 1:
        return pm[:b, :b]
    if i == 2:
        return pm[:b, b:]
    if i == 3:
        return pm[b:, :b]
    if i == 4:
        return pm[b:, b:]
    else:
        return 0


def matrix_paste(c_big: np.array, c_small: np.array) -> list:
    # Compute distance matrix between refined and dummy grid.
    dm = distance_matrix(c_big, c_small)
    inds = [
        np.unravel_index(np.argmin(dm[i], axis=None), dm[i].shape)[0]
        for i in range(dm.shape[0])
    ]
    return inds



#  Copyright (c) 2021. Robin Thibaut, Ghent University
import os
from os.path import join as jp

import numpy as np

from sklearn.cross_decomposition import CCA
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, PowerTransformer

from skbel.learning.bel import BEL


def init_bel():
    """
    Set all BEL pipelines
    """
    # Pipeline before CCA
    X_pre_processing = Pipeline(
        [
            ("scaler", StandardScaler(with_mean=False)),
            ("pca", PCA()),
        ]
    )
    Y_pre_processing = Pipeline(
        [
            ("scaler", StandardScaler(with_mean=False)),
            ("pca", PCA()),
        ]
    )

    # Canonical Correlation Analysis
    # Number of CCA components is chosen as the min number of PC
    n_pc_pred, n_pc_targ = (
        50,
        30,
    )
    cca = CCA(n_components=min(n_pc_targ, n_pc_pred), max_iter=500 * 20, tol=1e-6)

    # Pipeline after CCA
    X_post_processing = Pipeline(
        [("normalizer", PowerTransformer(method="yeo-johnson", standardize=True))]
    )
    Y_post_processing = Pipeline(
        [("normalizer", PowerTransformer(method="yeo-johnson", standardize=True))]
    )

    # Initiate BEL object
    bel_model = BEL(
        X_pre_processing=X_pre_processing,
        X_post_processing=X_post_processing,
        Y_pre_processing=Y_pre_processing,
        Y_post_processing=Y_post_processing,
        cca=cca,
    )

    # Set PC cut
    bel_model.X_n_pc = n_pc_pred
    bel_model.Y_n_pc = n_pc_targ

    return bel_model


def test_posterior():
    """Compare posterior mean and covariance with reference default values"""

    # Initiate BEL object
    bel = init_bel()
    # Set seed
    seed = 123456
    bel.seed = seed

    # Get file path
    my_path = os.path.dirname(os.path.abspath(__file__))

    X_train = np.load(jp(my_path, "X_train.npy"))
    y_train = np.load(jp(my_path, "y_train.npy"))

    X_test = np.load(jp(my_path, "X_test.npy"))

    # Fit
    bel.fit(X=X_train, Y=y_train)
    # Predict posterior mean and covariance
    post_mean, post_cov = bel.predict(X_test)

    # Compare with reference
    ref_mean = np.load(jp(my_path, "ref_mean.npy"))
    ref_covariance = np.load(jp(my_path, "ref_covariance.npy"))

    msg1 = "The posterior means are different"
    np.testing.assert_allclose(post_mean, ref_mean, atol=1e-6, err_msg=msg1)

    msg2 = "The posterior covariances are different"
    np.testing.assert_allclose(post_cov, ref_covariance, atol=1e-6, err_msg=msg2)



#  Copyright (c) 2021. Robin Thibaut, Ghent University

import math
import warnings

import numpy as np
import pandas as pd
from scipy import integrate, ndimage, stats
from sklearn.utils import check_array

from skbel.algorithms.extmath import get_block

__all__ = ["KDE", "kde_params", "posterior_conditional", "mvn_inference"]


class KDE:
    """
    Bivariate kernel density estimator.
    This class is adapted from the class of the same name in the package Seaborn 0.11.1
    https://seaborn.pydata.org/generated/seaborn.kdeplot.html
    """

    def __init__(
        self,
        *,
        bw_method: str = None,
        bw_adjust: float = 1,
        gridsize: int = 200,
        cut: float = 3,
        clip: list = None,
        cumulative: bool = False,
    ):
        """Initialize the estimator with its parameters.

        Parameters
        ----------
        bw_method : string, scalar, or callable, optional
            Method for determining the smoothing bandwidth to use; passed to
            :class:`scipy.stats.gaussian_kde`.
        bw_adjust : number, optional
            Factor that multiplicatively scales the value chosen using
            ``bw_method``. Increasing will make the curve smoother. See Notes.
        gridsize : int, optional
            Number of points on each dimension of the evaluation grid.
        cut : number, optional
            Factor, multiplied by the smoothing bandwidth, that determines how
            far the evaluation grid extends past the extreme datapoints. When
            set to 0, truncate the curve at the data limits.
        clip : pair of numbers None, or a pair of such pairs
            Do not evaluate the density outside of these limits.
        cumulative : bool, optional
            If True, estimate a cumulative distribution function.

        """
        if clip is None:
            clip = None, None

        self.bw_method = bw_method
        self.bw_adjust = bw_adjust
        self.gridsize = gridsize
        self.cut = cut
        self.clip = clip
        self.cumulative = cumulative

        self.support = None

    @staticmethod
    def _define_support_grid(
        x: np.array, bw: float, cut: float, clip: list, gridsize: int
    ):
        """Create the grid of evaluation points depending for vector x."""
        clip_lo = -np.inf if clip[0] is None else clip[0]
        clip_hi = +np.inf if clip[1] is None else clip[1]
        gridmin = max(x.min() - bw * cut, clip_lo)
        gridmax = min(x.max() + bw * cut, clip_hi)
        return np.linspace(gridmin, gridmax, gridsize)

    def _define_support_univariate(self, x: np.array, weights: np.array):
        """Create a 1D grid of evaluation points."""
        kde = self._fit(x, weights)
        bw = np.sqrt(kde.covariance.squeeze())
        grid = self._define_support_grid(x, bw, self.cut, self.clip, self.gridsize)
        return grid

    def _define_support_bivariate(self, x1: np.array, x2: np.array, weights: np.array):
        """Create a 2D grid of evaluation points."""
        clip = self.clip
        if clip[0] is None or np.isscalar(clip[0]):
            clip = (clip, clip)

        kde = self._fit([x1, x2], weights)
        bw = np.sqrt(np.diag(kde.covariance).squeeze())

        grid1 = self._define_support_grid(x1, bw[0], self.cut, clip[0], self.gridsize)
        grid2 = self._define_support_grid(x2, bw[1], self.cut, clip[1], self.gridsize)

        return grid1, grid2

    def define_support(
        self,
        x1: np.array,
        x2: np.array = None,
        weights: np.array = None,
        cache: bool = True,
    ):
        """Create the evaluation grid for a given data set."""
        if x2 is None:
            support = self._define_support_univariate(x1, weights)
        else:
            support = self._define_support_bivariate(x1, x2, weights)

        if cache:
            self.support = support

        return support

    def _fit(self, fit_data: np.array, weights: np.array = None):
        """Fit the scipy kde"""
        fit_kws = {"bw_method": self.bw_method}
        if weights is not None:
            fit_kws["weights"] = weights

        kde = stats.gaussian_kde(fit_data, **fit_kws)
        kde.set_bandwidth(kde.factor * self.bw_adjust)

        return kde

    def _eval_univariate(self, x: np.array, weights=None):
        """Fit and evaluate on univariate data."""
        support = self.support
        if support is None:
            support = self.define_support(x, cache=True)

        kde = self._fit(x, weights)

        if self.cumulative:
            s_0 = support[0]
            density = np.array([kde.integrate_box_1d(s_0, s_i) for s_i in support])
        else:
            density = kde(support)

        return density, support

    def _eval_bivariate(self, x1: np.array, x2: np.array, weights: np.array = None):
        """Fit and evaluate on bivariate data."""
        support = self.support
        if support is None:
            support = self.define_support(x1, x2, cache=False)

        kde = self._fit([x1, x2], weights)

        if self.cumulative:
            grid1, grid2 = support
            density = np.zeros((grid1.size, grid2.size))
            p0 = min(grid1), min(grid2)
            for i, xi in enumerate(grid1):
                for j, xj in enumerate(grid2):
                    density[i, j] = kde.integrate_box(p0, (xi, xj))

        else:
            xx1, xx2 = np.meshgrid(*support)
            density = kde([xx1.ravel(), xx2.ravel()]).reshape(xx1.shape)

        return density, support

    def __call__(self, x1, x2=None, weights=None):
        """Fit and evaluate on univariate or bivariate data."""
        if x2 is None:
            return self._eval_univariate(x1, weights)
        else:
            return self._eval_bivariate(x1, x2, weights)


def _univariate_density(
    data_variable: pd.DataFrame,
    estimate_kws: dict,
):
    # Initialize the estimator object
    estimator = KDE(**estimate_kws)

    all_data = data_variable.dropna()

    # Extract the data points from this sub set and remove nulls
    sub_data = all_data.dropna()
    observations = sub_data[data_variable]

    observation_variance = observations.var()
    if math.isclose(observation_variance, 0) or np.isnan(observation_variance):
        msg = "Dataset has 0 variance; skipping density estimate."
        warnings.warn(msg, UserWarning)

    # Estimate the density of observations at this level
    density, support = estimator(observations, weights=None)

    return density, support


def _bivariate_density(
    data: pd.DataFrame,
    estimate_kws: dict,
):
    """
    Estimate bivariate KDE
    :param data: DataFrame containing (x, y) data
    :param estimate_kws: KDE parameters
    :return:
    """

    estimator = KDE(**estimate_kws)

    all_data = data.dropna()

    # Extract the data points from this sub set and remove nulls
    sub_data = all_data.dropna()
    observations = sub_data[["x", "y"]]

    # Check that KDE will not error out
    variance = observations[["x", "y"]].var()
    if any(math.isclose(x, 0) for x in variance) or variance.isna().any():
        msg = "Dataset has 0 variance; skipping density estimate."
        warnings.warn(msg, UserWarning)

    # Estimate the density of observations at this level
    observations = observations["x"], observations["y"]
    density, support = estimator(*observations, weights=None)

    # Transform the support grid back to the original scale
    xx, yy = support

    support = xx, yy

    return density, support


def kde_params(
    x: np.array = None,
    y: np.array = None,
    bw: float = None,
    gridsize: int = 200,
    cut: float = 3,
    clip=None,
    cumulative: bool = False,
    bw_method: str = "scott",
    bw_adjust: int = 1,
):
    """
    Obtain density and support (grid) of the bivariate KDE
    :param x:
    :param y:
    :param bw:
    :param gridsize:
    :param cut:
    :param clip:
    :param cumulative:
    :param bw_method:
    :param bw_adjust:
    :return:
    """

    data = {"x": x, "y": y}
    frame = pd.DataFrame(data=data)

    # Handle deprecation of `bw`
    if bw is not None:
        bw_method = bw

    # Pack the kwargs for statistics.KDE
    estimate_kws = dict(
        bw_method=bw_method,
        bw_adjust=bw_adjust,
        gridsize=gridsize,
        cut=cut,
        clip=clip,
        cumulative=cumulative,
    )

    if y is None:
        density, support = _univariate_density(
            data_variable=frame, estimate_kws=estimate_kws
        )

    else:
        density, support = _bivariate_density(
            data=frame,
            estimate_kws=estimate_kws,
        )

    return density, support


def _pixel_coordinate(line: list, x_1d: np.array, y_1d: np.array):
    """
    Gets the pixel coordinate of the value x or y, in order to get posterior conditional probability given a KDE.
    :param line: Coordinates of the line we'd like to sample along [(x1, y1), (x2, y2)]
    :param x_1d: List of x coordinates along the axis
    :param y_1d: List of y coordinates along the axis
    :return:
    """
    # https://stackoverflow.com/questions/18920614/plot-cross-section-through-heat-map
    # Convert the line to pixel/index coordinates
    x_world, y_world = np.array(list(zip(*line)))
    col = y_1d.shape * (x_world - min(x_1d)) / x_1d.ptp()
    row = x_1d.shape * (y_world - min(y_1d)) / y_1d.ptp()

    # Interpolate the line at "num" points...
    num = 200
    row, col = [np.linspace(item[0], item[1], num) for item in [row, col]]

    return row, col


def _conditional_distribution(
    kde_array: np.array,
    x_array: np.array,
    y_array: np.array,
    x: float = None,
    y: float = None,
):
    """
    Compute the conditional posterior distribution p(x_array|y_array) given x or y.
    Provide only one observation ! Either x or y.
    Perform a cross-section in the KDE along the y axis.
    :param x: Observed data (horizontal axis)
    :param y: Observed data (vertical axis)
    :param kde_array: KDE of the prediction
    :param x_array: X grid (1D)
    :param y_array: Y grid (1D)
    :return:
    """

    # Coordinates of the line we'd like to sample along
    if x is not None:
        line = [(x, min(y_array)), (x, max(y_array))]
    elif y is not None:
        line = [(min(x_array), y), (max(x_array), y)]
    else:
        msg = "No observation point included."
        warnings.warn(msg, UserWarning)
        return 0

    # Convert line to row/column
    row, col = _pixel_coordinate(line=line, x_1d=x_array, y_1d=y_array)

    # Extract the values along the line, using cubic interpolation
    zi = ndimage.map_coordinates(kde_array, np.vstack((row, col)))

    return zi


def _normalize_distribution(post: np.array, support: np.array):
    """
    When a cross-section is performed along a bivariate KDE, the integral might not = 1.
    This function normalizes such functions so that their integral = 1.
    :param post: Values of the KDE cross-section
    :param support: Corresponding support
    :return:
    """
    a = integrate.simps(y=np.abs(post), x=support)

    if np.abs(a - 1) > 1e-4:  # Rule of thumb
        post *= 1 / a

    return post


def posterior_conditional(
    X: np.array, Y: np.array, X_obs: float = None, Y_obs: float = None
):
    """
    Computes the posterior distribution p(y|x_obs) or p(x|y_obs) by doing a cross section of the KDE of (d, h).
    :param X: Predictor (x-axis)
    :param Y: Target (y-axis)
    :param X_obs: Observation (predictor, x-axis)
    :param Y_obs: Observation (target, y-axis)
    :return:
    """
    # TODO: Adapt this method to evaluate the posterior distribution moments

    # Compute KDE
    dens, support = kde_params(x=X, y=Y)
    # Grid parameters
    xg, yg = support

    if X_obs is not None:
        support = yg
        # Extract the density values along the line, using cubic interpolation
        if type(X_obs) is list or tuple:
            X_obs = X_obs[0]
        post = _conditional_distribution(
            x=X_obs, x_array=xg, y_array=yg, kde_array=dens
        )
    elif Y_obs is not None:
        support = xg
        # Extract the density values along the line, using cubic interpolation
        if type(Y_obs) is list or tuple:
            Y_obs = X_obs[0]
        post = _conditional_distribution(
            y=Y_obs, x_array=xg, y_array=yg, kde_array=dens
        )

    else:
        msg = "No observation point included."
        warnings.warn(msg, UserWarning)
        return 0

    post = _normalize_distribution(post, support)

    # pmean = np.mean(support[~np.isclose(hp, 0)])
    # pvar = np.var(support[~np.isclose(hp, 0)])

    return post, support


def mvn_inference(
    X: np.array, Y: np.array, X_obs: np.array, **kwargs
) -> (np.array, np.array):
    """
    Estimating posterior mean and covariance of the target.
    .. [1] A. Tarantola. Inverse Problem Theory and Methods for Model Parameter Estimation.
           SIAM, 2005. Pages: 70-71
    :param X: Canonical Variate of the training data
    :param Y: Canonical Variate of the training target, gaussian-distributed
    :param X_obs: Canonical Variate of the observation
    :return: y_posterior_mean, y_posterior_covariance
    """

    Y = check_array(Y, copy=True, ensure_2d=False)
    X = check_array(X, copy=True, ensure_2d=False)
    X_obs = check_array(X_obs, copy=True, ensure_2d=False)

    # Size of the set
    n_training = X.shape[0]

    # Computation of the posterior mean in Canonical space
    y_mean = np.mean(Y, axis=0)  # (n_comp_CCA, 1)
    # Mean is 0, as expected.
    y_mean = np.where(np.abs(y_mean) < 1e-12, 0, y_mean)

    # Evaluate the covariance in h (in Canonical space)
    # Very close to the Identity matrix
    # (n_comp_CCA, n_comp_CCA)
    y_cov = np.cov(Y.T)

    if "x_cov" in kwargs.keys():
        x_cov = kwargs["x_cov"]
    else:
        x_cov = np.zeros(shape=y_cov.shape)

    # Linear modeling h to d (in canonical space) with least-square criterion.
    # Pay attention to the transpose operator.
    # Computes the vector g that approximately solves the equation y @ g = x.
    g = np.linalg.lstsq(Y, X, rcond=None)[0].T
    # Replace values below threshold by 0.
    g = np.where(np.abs(g) < 1e-12, 0, g)  # (n_comp_CCA, n_comp_CCA)

    # Modeling error due to deviations from theory
    # (n_components_CCA, n_training)
    x_ls_predicted = np.matmul(Y, g.T)
    x_modeling_mean_error = np.mean(X - x_ls_predicted, axis=0)  # (n_comp_CCA, 1)
    x_modeling_error = (
        X - x_ls_predicted - np.tile(x_modeling_mean_error, (n_training, 1))
    )
    # (n_comp_CCA, n_training)

    # Information about the covariance of the posterior distribution in Canonical space.
    x_modeling_covariance = np.cov(x_modeling_error.T)  # (n_comp_CCA, n_comp_CCA)

    # Build block matrix
    s11 = y_cov
    s12 = y_cov @ g.T
    s21 = g @ y_cov
    s22 = g @ y_cov @ g.T + x_cov + x_modeling_covariance
    block = np.block([[s11, s12], [s21, s22]])

    # Inverse
    delta = np.linalg.pinv(block)
    # Partition block
    d11 = get_block(delta, 1)
    d12 = get_block(delta, 2)

    # Observe that posterior covariance does not depend on observed x.
    y_posterior_covariance = np.linalg.pinv(d11)  # (n_comp_CCA, n_comp_CCA)
    # Computing the posterior mean is simply a linear operation, given precomputed posterior covariance.
    y_posterior_mean = y_posterior_covariance @ (
        d11 @ y_mean - d12 @ (X_obs[0] - x_modeling_mean_error - y_mean @ g.T)
    )  # (n_comp_CCA,)

    return y_posterior_mean, y_posterior_covariance



"""Collection of functions to use the PCA class from scikit-learn."""

#  Copyright (c) 2021. Robin Thibaut, Ghent University


import numpy as np
import pandas as pd
from loguru import logger
from sklearn.decomposition import PCA
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from skbel import utils

__all__ = ["PC"]


class PC:
    def __init__(
        self,
        name: str,
        training_df: pd.DataFrame = None,
        test_df: pd.DataFrame = None,
        directory: str = None,
    ):
        """
        Given a set of training data and one observation (optional), performs necessary dimension reduction
        and transformations.
        :param name: str: name of the dataset (e.g. 'data', 'target'...)
        :param training_df: pd.DataFrame Training set
        :param test_df: pd.DataFrame: Test set
        :param directory: str: Path to the folder in which to save the pickle
        """
        logger.info("Initiating dimension reduction")
        self.directory = directory
        self.name = name  # str, name of the object

        self.training_df = training_df
        logger.info(f"Training set {self.training_df.shape}")
        self.test_df = test_df

        try:
            logger.info(f"Test set {self.test_df.shape}")
        except AttributeError:
            logger.info("No test set")

        self.training_pc_df = None
        self.test_pc_df = None

        # Divide the sample by their standard deviation.
        self.scaler = StandardScaler(with_mean=False)
        # The samples are automatically scaled by scikit-learn PCA()
        self.operator = PCA()  # PCA operator (scikit-learn instance)
        self.pipe = make_pipeline(self.scaler, self.operator, verbose=False)

        self.n_samples = self.training_df.shape[0]
        self.n_pc_cut = None  # Number of components to keep

        # Number of training samples
        self.training_pc = None  # Training PCA scores

    def training_fit_transform(self):
        """
        Instantiate the PCA object and transforms training data to scores.
        :return: numpy.array: PC training
        """

        logger.info("Fitting and transforming training data")
        self.pipe.fit(self.training_df)
        training_pc = self.pipe.transform(self.training_df.to_numpy())

        # Store PC dataframe
        self.training_pc_df = utils.i_am_framed(
            array=training_pc, ids=self.training_df.index
        )

        return training_pc

    def test_transform(self, test_roots: list = None):
        """
        Transforms observation to PC scores.
        :param test_roots: list: List containing observation id (str)
        :return: numpy.array: Observation PC
        """

        if test_roots:
            selection = self.test_df.loc[test_roots]
            ids = test_roots
        else:
            selection = self.test_df.to_numpy()
            ids = self.test_df.index

        logger.info("Fitting and transforming test data")
        # Transform prediction data into principal components
        pc_prediction = self.pipe.transform(selection)

        # Store PC dataframe
        self.test_pc_df = utils.i_am_framed(array=pc_prediction, ids=ids)

        return pc_prediction

    def perc_2_comp(self, perc: float) -> int:
        """
        Given an explained variance percentage, returns the number of components
        necessary to obtain that level.
        :param perc: float: Percentage between 0 and 1
        """
        evr = np.cumsum(self.operator.explained_variance_ratio_)
        self.n_pc_cut = len(np.where(evr <= perc)[0])

        return self.n_pc_cut

    def perc_comp(self, n_c: int) -> float:
        """
        Returns the explained variance percentage given a number of components n_c.
        :param n_c: int: Number of components to keep
        """
        evr = np.cumsum(self.operator.explained_variance_ratio_)

        return evr[n_c - 1]

    def comp_refresh(self, n_comp: int = None) -> (np.array, np.array):
        """
        Given a number of components to keep, returns the PC array with the corresponding shape.
        :param n_comp: int: Number of components
        :return:
        """

        if n_comp is not None:
            self.n_pc_cut = (
                n_comp  # Assign the number of components in the class for later use
            )

        # Reloads the original training components
        pc_training = self.training_pc_df.to_numpy()
        pc_training = pc_training[:, : self.n_pc_cut]  # Cut

        pc_prediction = self.test_pc_df

        if self.test_pc_df is not None:
            pc_prediction = (
                self.test_pc_df.to_numpy()
            )  # Reloads the original test components
            pc_prediction = pc_prediction[:, : self.n_pc_cut]  # Cut

        return pc_training, pc_prediction

    def random_pc(self, n_rand: int):
        """
        Randomly selects PC components from the original training matrix.
        :param n_rand: int: Number of random PC to use
        :return numpy.array: Random PC scores
        """
        rand_rows = np.random.choice(
            self.n_samples, n_rand
        )  # Selects n_posts rows from the training array
        # Extracts those rows, from the number of
        score_selection = self.training_pc_df[rand_rows][self.n_pc_cut :]
        # components used until the end of the array.

        # For each column of shape n_samples, n_components, selects a random PC component to add.
        test = [
            np.random.choice(score_selection[:][i])
            for i in range(score_selection.shape[1])
        ]

        return np.array(test)

    def custom_inverse_transform(self, pc_to_invert: np.array, n_comp: int = None):
        """
        Inverse fit_transform PC based on the desired number of PC (stored in the shape of the argument).
        The self.operator.components contains all components.
        :param pc_to_invert: np.array: (n_samples, n_components) PC array to back-fit_transform to physical space
        :param n_comp: int: Number of components to back-fit_transform with
        :return: numpy.array: Back transformed array
        """
        if n_comp is None:
            n_comp = self.n_pc_cut

        # TODO: (optimization) only fit after dimension check
        op_cut = make_pipeline(self.scaler, PCA(n_components=n_comp))
        op_cut.fit(self.training_df)

        inv = op_cut.inverse_transform(pc_to_invert[:, :n_comp])

        return inv



"""Bayesian Evidential Learning Framework"""

#  Copyright (c) 2021. Robin Thibaut, Ghent University

import numpy as np
from sklearn.base import (
    BaseEstimator,
    TransformerMixin,
    MultiOutputMixin,
)
from sklearn.utils import check_array
from sklearn.utils.validation import (
    check_is_fitted,
    check_consistent_length,
)

from ..algorithms import mvn_inference, posterior_conditional
from ..algorithms._statistics import _normalize_distribution


class BEL(TransformerMixin, MultiOutputMixin, BaseEstimator):
    """
    Heart of the framework. Inherits from scikit-learn base classes.
    """

    def __init__(
        self,
        mode: str = "mvn",
        copy: bool = True,
        *,
        X_pre_processing=None,
        Y_pre_processing=None,
        X_post_processing=None,
        Y_post_processing=None,
        cca=None,
    ):
        """
        :param mode: How to infer the posterior distribution. "mvn" (default) or "kde"
        :param copy: Whether to copy arrays or not (default is True).
        :param X_pre_processing: sklearn pipeline for pre-processing the predictor.
        :param Y_pre_processing: sklearn pipeline for pre-processing the target.
        :param X_post_processing: sklearn pipeline for post-processing the predictor.
        :param X_post_processing: sklearn pipeline for post-processing the target.
        :param cca: sklearn cca object
        """
        self.copy = copy
        # How to infer the posterior parameters
        self.mode = mode

        # Processing pipelines
        self.X_pre_processing = X_pre_processing
        self.Y_pre_processing = Y_pre_processing
        self.X_post_processing = X_post_processing
        self.Y_post_processing = Y_post_processing
        self.cca = cca

        # Posterior parameters
        self.posterior_mean = None
        self.posterior_covariance = None

        # Parameters for sampling
        self._seed = None
        self._n_posts = None

        # Original dataset
        self.X_shape, self.Y_shape = None, None
        self.X, self.Y = None, None
        self.X_obs, self.Y_obs = None, None  # Observation data

        # Dataset after preprocessing (dimension-reduced by self.X_n_pc, self.Y_n_pc)
        self._X_n_pc, self._Y_n_pc = None, None
        self.X_pc, self.Y_pc = None, None
        self.X_obs_pc, self.Y_obs_pc = None, None
        # Dataset after learning
        self.X_c, self.Y_c = None, None
        self.X_obs_c, self.Y_obs_c = None, None
        # Dataset after postprocessing
        self.X_f, self.Y_f = None, None
        self.X_obs_f, self.Y_obs_f = None, None

    # The following properties are central to the BEL framework
    @property
    def X_n_pc(self):
        """Number of components to keep after pre-processing (dimensionality reduction)"""
        return self._X_n_pc

    @X_n_pc.setter
    def X_n_pc(self, X_n_pc):
        self._X_n_pc = X_n_pc

    @property
    def Y_n_pc(self):
        """Number of components to keep after pre-processing (dimensionality reduction)"""
        return self._Y_n_pc

    @Y_n_pc.setter
    def Y_n_pc(self, Y_n_pc):
        self._Y_n_pc = Y_n_pc

    @property
    def n_posts(self):
        """Number of sample to extract from the posterior multivariate distribution after post-processing"""
        return self._n_posts

    @n_posts.setter
    def n_posts(self, n_posts):
        self._n_posts = n_posts

    @property
    def seed(self):
        """Seed a.k.a. random state to reproduce the same samples"""
        return self._seed

    @seed.setter
    def seed(self, seed):
        self._seed = seed
        np.random.seed(self._seed)

    def fit(self, X, Y):
        """
        Fit all pipelines.
        Requires the predictor and target to be panda dataframes with a "physical_shape" attribute.
        The "physical_shape" is simply the original dimension of the data.
        :param X: Predictor dataframe.
        :param Y: Target dataframe.
        :return:
        """
        check_consistent_length(X, Y)
        self.X, self.Y = X, Y  # Save dataframe with names
        # Store original shape
        try:
            self.X_shape, self.Y_shape = (
                X.attrs["physical_shape"],
                Y.attrs["physical_shape"],
            )
        except AttributeError:
            self.X_shape, self.Y_shape = (1,), (1,)

        _X = self._validate_data(
            X, dtype=np.float64, copy=self.copy, ensure_min_samples=2
        )
        _Y = check_array(Y, dtype=np.float64, copy=self.copy, ensure_2d=False)

        _xt, _yt = (
            self.X_pre_processing.fit_transform(_X),
            self.Y_pre_processing.fit_transform(_Y),
        )

        _xt, _yt = (
            _xt[:, : self.X_n_pc],
            _yt[:, : self.Y_n_pc],
        )  # Cut PC

        # Dataset after preprocessing
        self.X_pc, self.Y_pc = _xt, _yt

        # Canonical variates
        _xc, _yc = self.cca.fit_transform(X=_xt, y=_yt)

        self.X_c, self.Y_c = _xc, _yc

        # CV Normalized
        _xf, _yf = (
            self.X_post_processing.fit_transform(_xc),
            self.Y_post_processing.fit_transform(_yc),
        )

        self.X_f, self.Y_f = _xf, _yf

        return self

    def transform(self, X=None, Y=None) -> (np.array, np.array):
        """
        Transform data across all pipelines
        :param X: Predictor array.
        :param Y: Target array.
        :return: Post-processed variables
        """

        check_is_fitted(self.cca)
        # The key here is to cut PC's based on the number defined in configuration file

        if X is not None and Y is None:
            X = check_array(X, copy=True)
            _xt = self.X_pre_processing.transform(X)
            _xt = _xt[:, : self.X_n_pc]
            _xc = self.cca.transform(X=_xt)
            _xp = self.X_post_processing.transform(_xc)

            return _xp

        elif Y is not None and X is None:
            Y = check_array(Y, copy=True, ensure_2d=False)
            _xt, _yt = (
                self.X_pre_processing.transform(self.X),
                self.Y_pre_processing.transform(Y),
            )
            _xt, _yt = (
                _xt[:, : self.X_n_pc],
                _yt[:, : self.Y_n_pc],
            )
            _, _yc = self.cca.transform(X=_xt, Y=_yt)
            _yp = self.Y_post_processing.transform(_yc)

            return _yp

        else:
            _xt, _yt = (
                self.X_pre_processing.transform(self.X),
                self.Y_pre_processing.transform(self.Y),
            )
            _xt, _yt = (
                _xt[:, : self.X_n_pc],
                _yt[:, : self.Y_n_pc],
            )
            _xc, _yc = self.cca.transform(X=_xt, Y=_yt)

            _xp, _yp = (
                self.X_post_processing.transform(_xc),
                self.Y_post_processing.transform(_yc),
            )

            return _xp, _yp

    def random_sample(self, n_posts: int = None) -> np.array:
        """
        Random sample the inferred posterior Gaussian distribution
        :param n_posts:
        :return:
        """
        # Set the seed for later use
        if self.seed is None:
            self.seed = np.random.randint(2 ** 32 - 1, dtype="uint32")

        check_is_fitted(self.cca)
        if n_posts is None:
            n_posts = self.n_posts
        else:
            self.n_posts = n_posts
        # Draw n_posts random samples from the multivariate normal distribution :
        # Pay attention to the transpose operator
        np.random.seed(self.seed)
        Y_samples = np.random.multivariate_normal(
            mean=self.posterior_mean, cov=self.posterior_covariance, size=n_posts
        )

        return Y_samples

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fit-Transform across all pipelines
        :param X:
        :param y:
        :return:
        """

        return self.fit(X, y).transform(X, y)

    def predict(self, X_obs) -> (np.array, np.array):
        """
        Make predictions, in the BEL fashion.
        """
        self.X_obs = X_obs  # Save dataframe with name
        try:
            X_obs = check_array(self.X_obs)
        except ValueError:
            X_obs = check_array(self.X_obs.to_numpy().reshape(1, -1))
        # Project observed data into canonical space.
        X_obs = self.X_pre_processing.transform(X_obs)
        X_obs = X_obs[:, : self.X_n_pc]
        self.X_obs_pc = X_obs
        X_obs = self.cca.transform(X_obs)
        self.X_obs_c = X_obs
        X_obs = self.X_post_processing.transform(X_obs)
        self.X_obs_f = X_obs

        # Evaluate the covariance in d (here we assume no data error, so C is identity times a given factor)
        # Number of PCA components for the curves
        x_dim = self.X_n_pc
        noise = 0.01
        # I matrix. (n_comp_PCA, n_comp_PCA)
        x_cov = np.eye(x_dim) * noise
        # (n_comp_CCA, n_comp_CCA)
        # Get the rotation matrices
        x_rotations = self.cca.x_rotations_
        x_cov = x_rotations.T @ x_cov @ x_rotations
        dict_args = {"x_cov": x_cov}

        X, Y = self.X_f, self.Y_f
        # Estimate the posterior mean and covariance
        if self.mode == "mvn":
            self.posterior_mean, self.posterior_covariance = mvn_inference(
                X=X,
                Y=Y,
                X_obs=X_obs,
                **dict_args,
            )
        else:
            # KDE inference
            for comp_n in range(self.cca.n_components):
                # Conditional:
                hp, sup = posterior_conditional(
                    X=self.X_f.T[comp_n], Y=self.Y_f.T[comp_n], X_obs=self.X_obs_f.T[comp_n]
                )
                hp[np.abs(hp) < 1e-12] = 0  # Set very small values to 0.
                hp = _normalize_distribution(hp, sup)
                if comp_n > 0:
                    my_arr = np.concatenate((my_arr, [hp]), axis=0)
                    my_sup = np.concatenate((my_sup, [sup]), axis=0)
                else:
                    my_arr = [hp]
                    my_sup = [sup]

                mean = sum(sup * hp) / sum(hp)  # Find mean
                sigma = np.sqrt(sum(hp * (sup - mean) ** 2) / sum(hp))  # Find std dev
                s = np.random.normal(mean, sigma, 200)  # Sample

                if comp_n > 0:
                    smean = np.concatenate((smean, [mean]), axis=0)
                    sbig = np.concatenate((sbig, [s]), axis=0)
                else:
                    smean = [mean]
                    sbig = [s]

            self.posterior_mean = smean

            self.posterior_covariance = np.cov(sbig)

        return self.posterior_mean, self.posterior_covariance

    def inverse_transform(
        self,
        Y_pred,
    ) -> np.array:
        """
        Back-transforms the sampled gaussian distributed posterior Y to their physical space.
        :param Y_pred:
        :return: forecast_posterior
        """
        check_is_fitted(self.cca)
        Y_pred = check_array(Y_pred)

        y_post = self.Y_post_processing.inverse_transform(
            Y_pred
        )  # Posterior CCA scores
        y_post = (
            np.matmul(y_post, self.cca.y_loadings_.T) * self.cca.y_std_
            + self.cca.y_mean_
        )  # Posterior PC scores

        # Back transform PC scores
        nc = self.Y_pre_processing["pca"].n_components_  # Number of components
        dummy = np.zeros((self.n_posts, nc))  # Create a dummy matrix filled with zeros
        dummy[
            :, : y_post.shape[1]
        ] = y_post  # Fill the dummy matrix with the posterior PC
        y_post = self.Y_pre_processing.inverse_transform(dummy)  # Inverse transform

        return y_post
